# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import textwrap

import pytest

from specfile.formatter import format_expression


@pytest.mark.parametrize(
    "original, reformatted",
    [
        ("func('arg')", "func('arg')"),
        ("func('arg1', \"arg2\")", "func('arg1', 'arg2')"),
        ("func('arg1', 'arg\\'2')", "func('arg1', \"arg'2\")"),
        ("func('arg', kwarg='val')", "func('arg', kwarg='val')"),
        ("(None)", "None"),
        ("(None,)", "(None,)"),
        ("('a',\"b\",3)", "('a', 'b', 3)"),
        ("[1,2,3,4]", "[1, 2, 3, 4]"),
        ("{'key':'val'}", "{'key': 'val'}"),
        ("<ENUM_ITEM_1: 1>", "<ENUM_ITEM_1: 1>"),
        (
            "func1('first argument', True, func2(kwarg={42: ['nested list item 1', "
            "'nested list item 2', 'nested list item 3']}), 0, indent='    ', "
            "spec=<rpm.spec object at 0x7fe1ae1a6b30>)",
            textwrap.dedent(
                """\
                func1(
                    'first argument',
                    True,
                    func2(
                        kwarg={
                            42: [
                                'nested list item 1',
                                'nested list item 2',
                                'nested list item 3',
                            ],
                        },
                    ),
                    0,
                    indent='    ',
                    spec=<rpm.spec object at 0x7fe1ae1a6b30>,
                )"""
            ),
        ),
    ],
)
def test_format_expression(original, reformatted):
    assert format_expression(original) == reformatted



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import copy

import pytest

from specfile.sections import Section
from specfile.tags import Comment, Comments, Tag, Tags


def test_find():
    tags = Tags(
        [
            Tag("Name", "test", ": ", Comments()),
            Tag("Version", "0.1", ": ", Comments()),
            Tag("Release", "1%{?dist}", ": ", Comments()),
        ]
    )
    assert tags.find("version") == 1
    with pytest.raises(ValueError):
        tags.find("Epoch")


def test_parse():
    tags = Tags.parse(
        Section(
            "package",
            data=[
                "%{?scl:%scl_package scltest}",
                "",
                "%global ver_major 1",
                "%global ver_minor 0",
                "",
                "# this is a test package",
                "# not to be used in production",
                "Name:    test",
                "Version: %{ver_major}.%{ver_minor}",
                "  # this is a valid comment",
                "Release: 1%{?dist}",
                "",
                "%if 0",
                "Epoch:   1",
                "%endif",
                "",
                "Requires:          make",
                "Requires(post):    bash",
                "",
                "Provides:          testX = %{version}-%{release}",
                "Provides:          bundled(superlib) = 2.42.0",
                "",
                "%{?fedora:Suggests:          diffutils}",
            ],
        )
    )
    assert tags[0].comments._preceding_lines[0] == "%{?scl:%scl_package scltest}"
    assert tags[0].name == "Name"
    assert tags[0].comments[0].text == "this is a test package"
    assert tags[0].comments[1].text == "not to be used in production"
    assert tags[1].name == "Version"
    assert tags[1].value == "%{ver_major}.%{ver_minor}"
    assert not tags[1].comments
    assert tags.release.comments[0].prefix == "  # "
    assert tags.epoch.name == "Epoch"
    assert tags.requires.value == "make"
    assert "requires(post)" in tags
    assert tags[-4].name == "Requires(post)"
    assert tags[-3].name == "Provides"
    assert tags[-3].value == "testX = %{version}-%{release}"
    assert tags[-2].name == "Provides"
    assert tags[-2].value == "bundled(superlib) = 2.42.0"
    assert tags[-1].name == "Suggests"
    assert tags.suggests.value == "diffutils"


def test_get_raw_section_data():
    tags = Tags(
        [
            Tag(
                "Name",
                "test",
                ":    ",
                Comments(
                    [
                        Comment("this is a test package"),
                        Comment("not to be used in production"),
                    ],
                    [
                        "%{?scl:%scl_package scltest}",
                        "",
                        "%global ver_major 1",
                        "%global ver_minor 0",
                        "",
                    ],
                ),
            ),
            Tag("Version", "%{ver_major}.%{ver_minor}", ": ", Comments()),
            Tag(
                "Release",
                "1%{?dist}",
                ": ",
                Comments([Comment("this is a valid comment", "  # ")]),
            ),
            Tag("Epoch", "1", ":   ", Comments([], ["", "%if 0"])),
            Tag("Requires", "make", ":          ", Comments([], ["%endif", ""])),
            Tag("Requires(post)", "bash", ":    ", Comments()),
            Tag(
                "Suggests",
                "diffutils",
                ":          ",
                Comments([], [""]),
                "%{?fedora:",
                "}",
            ),
        ],
        [],
    )
    assert tags.get_raw_section_data() == [
        "%{?scl:%scl_package scltest}",
        "",
        "%global ver_major 1",
        "%global ver_minor 0",
        "",
        "# this is a test package",
        "# not to be used in production",
        "Name:    test",
        "Version: %{ver_major}.%{ver_minor}",
        "  # this is a valid comment",
        "Release: 1%{?dist}",
        "",
        "%if 0",
        "Epoch:   1",
        "%endif",
        "",
        "Requires:          make",
        "Requires(post):    bash",
        "",
        "%{?fedora:Suggests:          diffutils}",
    ]


def test_copy_tags():
    tags = Tags(
        [
            Tag("Name", "test", ": ", Comments()),
        ]
    )
    shallow_copy = copy.copy(tags)
    assert shallow_copy == tags
    assert shallow_copy is not tags
    assert shallow_copy[0] is tags[0]
    deep_copy = copy.deepcopy(tags)
    assert deep_copy == tags
    assert deep_copy is not tags
    assert deep_copy[0] is not tags[0]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest

from specfile.specfile import Specfile


@pytest.mark.parametrize(
    "raw_release, release, dist, minorbump",
    [
        ("1%{?dist}", "1", "%{?dist}", None),
        ("0.1%{prerelease}%{dist}", "0.1%{prerelease}", "%{dist}", None),
        ("%{_short_release}%{?dist}.20", "%{_short_release}", "%{?dist}", 20),
        ("28%dist", "28", "%dist", None),
        ("%{release_string}", "%{release_string}", None, None),
    ],
)
def test_split_raw_release(raw_release, release, dist, minorbump):
    assert Specfile._split_raw_release(raw_release) == (release, dist, minorbump)


@pytest.mark.parametrize(
    "raw_release, release, result",
    [
        ("1%{?dist}", "28", "28%{?dist}"),
        ("0.1%{prerelease}%{dist}", "1", "1%{dist}"),
        ("%{_short_release}%{?dist}.20", "3", "3%{?dist}"),
        ("28%dist", "0.1", "0.1%dist"),
        ("%{release_string}", "1", "1"),
    ],
)
def test_get_updated_release(raw_release, release, result):
    assert Specfile._get_updated_release(raw_release, release) == result



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import ast
import functools
from typing import Callable

from specfile.exceptions import SpecfileException


def format_expression(expression: str, line_length_threshold: int = 80) -> str:
    """
    Formats the specified Python expression.

    Only supports a small subset of Python AST that should be sufficient for use in __repr__().

    Args:
        expression: Python expression to reformat.
        line_length_threshold: Threshold for line lengths. It's not a hard limit,
          it can be exceeded in some cases.

    Returns:
        Formatted expression.

    Raises:
        SyntaxError if the expression is not parseable.
        SpecfileException if there is an unsupported AST node in the expression.
    """

    def fmt(node, indent=0, prefix="", multiline=False):
        result = " " * indent + prefix
        if isinstance(node, ast.Constant):
            result += repr(node.value)
        elif isinstance(node, (ast.Tuple, ast.List, ast.Dict, ast.Call)):
            if isinstance(node, ast.Tuple):
                start, end = "(", ")" if multiline or len(node.elts) != 1 else ",)"
                items = [(None, e) for e in node.elts]
                delimiter = None
            elif isinstance(node, ast.List):
                start, end = "[", "]"
                items = [(None, e) for e in node.elts]
                delimiter = None
            elif isinstance(node, ast.Dict):
                start, end = "{", "}"
                items = [(fmt(k, 0), v) for k, v in zip(node.keys, node.values)]
                delimiter = ": "
            elif isinstance(node, ast.Call):
                start, end = f"{node.func.id}(", ")"
                items = [(None, a) for a in node.args] + [
                    (kw.arg, kw.value) for kw in node.keywords
                ]
                delimiter = "="
            result += start
            if multiline:
                result += "\n"
            while items:
                key, value = items.pop(0)
                result += fmt(
                    value,
                    indent + 4 if multiline else 0,
                    f"{key}{delimiter}" if key else "",
                )
                if multiline:
                    result += ",\n"
                elif items:
                    result += ", "
            if multiline:
                result += " " * indent
            result += end
        else:
            raise SpecfileException(
                f"Unsupported AST node: ast.{node.__class__.__name__}"
            )
        if not multiline and len(result) > line_length_threshold:
            result = fmt(node, indent, prefix, True)
        return result

    def find_matching_bracket(value, index):
        level = 0
        for i in range(index, len(value)):
            if value[i] == ">":
                level -= 1
                if level <= 0:
                    return i + 1
            elif value[i] == "<":
                level += 1
        return None

    placeholders = []
    while True:
        try:
            tree = ast.parse(expression, mode="eval")
        except SyntaxError as e:
            # some objects (e.g. enums) are not represented by a valid expression,
            # replace their representation with placeholders
            # (assume such a representation is enclosed in <> and doesn't spawn
            # across multiple lines)
            lines = expression.splitlines()
            if not e.lineno or not e.offset:
                raise
            if lines[e.lineno - 1][e.offset - 1] != "<":
                raise
            index = find_matching_bracket(lines[e.lineno - 1], e.offset - 1)
            if not index:
                raise
            value = lines[e.lineno - 1][e.offset - 1 : index]
            # convert the value to a string literal
            placeholder = repr(value)
            placeholders.append((placeholder, value))
            lines[e.lineno - 1] = (
                lines[e.lineno - 1][: e.offset - 1]
                + placeholder
                + lines[e.lineno - 1][index:]
            )
            expression = "\n".join(lines)
        else:
            break
    result = fmt(tree.body)
    # revert placeholders
    for placeholder, value in placeholders:
        result = result.replace(placeholder, value, 1)
    return result


def formatted(function: Callable[..., str]) -> Callable[..., str]:
    """Decorator for formatting the output of __repr__()."""

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        result = function(*args, **kwargs)
        try:
            return format_expression(result)
        except (SyntaxError, SpecfileException):
            return result

    return wrapper



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import copy

import pytest

from specfile.macro_definitions import MacroDefinition, MacroDefinitions


def test_find():
    macro_definitions = MacroDefinitions(
        [
            MacroDefinition("gitdate", "20160901", True, ("", " ", "     ", "")),
            MacroDefinition(
                "commit",
                "9ab9717cf7d1be1a85b165a8eacb71b9e5831113",
                True,
                ("", " ", "      ", ""),
            ),
            MacroDefinition(
                "shortcommit", "%(c=%{commit}; echo ${c:0:7})", True, ("", " ", " ", "")
            ),
        ]
    )
    assert macro_definitions.find("gitdate") == 0
    assert macro_definitions.find("shortcommit") == 2
    with pytest.raises(ValueError):
        macro_definitions.find("gittag")


def test_get():
    macro_definitions = MacroDefinitions(
        [
            MacroDefinition("gitdate", "20160901", True, ("", " ", "     ", "")),
            MacroDefinition(
                "commit",
                "9ab9717cf7d1be1a85b165a8eacb71b9e5831113",
                True,
                ("", " ", "      ", ""),
            ),
            MacroDefinition(
                "shortcommit", "%(c=%{commit}; echo ${c:0:7})", True, ("", " ", " ", "")
            ),
        ]
    )
    assert (
        macro_definitions.get("commit").body
        == "9ab9717cf7d1be1a85b165a8eacb71b9e5831113"
    )
    with pytest.raises(ValueError):
        macro_definitions.get("gittag")


def test_parse():
    macro_definitions = MacroDefinitions.parse(
        [
            "%global gitdate     20160901",
            "%global commit      9ab9717cf7d1be1a85b165a8eacb71b9e5831113",
            "%global shortcommit %(c=%{commit}; echo ${c:0:7})",
            "",
            "Name:           test",
            "Version:        0.1.0",
            "",
            "%define desc(x) Test spec file containing several \\",
            "macro definitions in various formats (%?1)",
            "",
            "%define example() %{expand:",
            "This an example of a macro definition with body ",
            "spawning across mutiple lines}",
        ]
    )
    assert macro_definitions[0].name == "gitdate"
    assert macro_definitions[1].name == "commit"
    assert macro_definitions.commit.body == "9ab9717cf7d1be1a85b165a8eacb71b9e5831113"
    assert macro_definitions[2].name == "shortcommit"
    assert macro_definitions[3].name == "desc(x)"
    assert macro_definitions[3].body == (
        "Test spec file containing several \\\n"
        "macro definitions in various formats (%?1)"
    )
    assert macro_definitions[-1].name == "example()"
    assert macro_definitions[-1].body == (
        "%{expand:\n"
        "This an example of a macro definition with body \n"
        "spawning across mutiple lines}"
    )


def test_get_raw_data():
    macro_definitions = MacroDefinitions(
        [
            MacroDefinition("gitdate", "20160901", True, ("", " ", "     ", "")),
            MacroDefinition(
                "commit",
                "9ab9717cf7d1be1a85b165a8eacb71b9e5831113",
                True,
                ("", " ", "      ", ""),
            ),
            MacroDefinition(
                "shortcommit", "%(c=%{commit}; echo ${c:0:7})", True, ("", " ", " ", "")
            ),
            MacroDefinition(
                "desc(x)",
                "Test spec file containing several \\\nmacro definitions in various formats (%?1)",
                False,
                ("", " ", " ", ""),
                [
                    "",
                    "Name:           test",
                    "Version:        0.1.0",
                    "",
                ],
            ),
            MacroDefinition(
                "example()",
                "%{expand:\n"
                "This an example of a macro definition with body \n"
                "spawning across mutiple lines}",
                False,
                ("", " ", " ", ""),
                [""],
            ),
        ]
    )
    assert macro_definitions.get_raw_data() == [
        "%global gitdate     20160901",
        "%global commit      9ab9717cf7d1be1a85b165a8eacb71b9e5831113",
        "%global shortcommit %(c=%{commit}; echo ${c:0:7})",
        "",
        "Name:           test",
        "Version:        0.1.0",
        "",
        "%define desc(x) Test spec file containing several \\",
        "macro definitions in various formats (%?1)",
        "",
        "%define example() %{expand:",
        "This an example of a macro definition with body ",
        "spawning across mutiple lines}",
    ]


def test_copy_macro_definitions():
    macro_definitions = MacroDefinitions(
        [
            MacroDefinition(
                "commit",
                "9ab9717cf7d1be1a85b165a8eacb71b9e5831113",
                True,
                ("", " ", "      ", ""),
            ),
        ],
    )
    shallow_copy = copy.copy(macro_definitions)
    assert shallow_copy == macro_definitions
    assert shallow_copy is not macro_definitions
    assert shallow_copy[0] is macro_definitions[0]
    deep_copy = copy.deepcopy(macro_definitions)
    assert deep_copy == macro_definitions
    assert deep_copy is not macro_definitions
    assert deep_copy[0] is not macro_definitions[0]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import itertools
import re
from abc import ABC
from string import Template
from typing import TYPE_CHECKING, Generator, List, Optional, Pattern, Set, Tuple

from specfile.exceptions import UnterminatedMacroException
from specfile.formatter import formatted
from specfile.macros import Macros

if TYPE_CHECKING:
    from specfile.specfile import Specfile

SUBSTITUTION_GROUP_PREFIX = "sub_"


class Node(ABC):
    """Base class for all nodes."""

    ...


class StringLiteral(Node):
    """Node representing string literal."""

    def __init__(self, value: str) -> None:
        self.value = value

    @formatted
    def __repr__(self) -> str:
        return f"StringLiteral({self.value!r})"

    def __str__(self) -> str:
        return self.value

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.value == other.value


class ShellExpansion(Node):
    """Node representing shell expansion, e.g. %(whoami)."""

    def __init__(self, body: str) -> None:
        self.body = body

    @formatted
    def __repr__(self) -> str:
        # determine class name dynamically so that inherited classes
        # don't have to reimplement __repr__()
        return f"{self.__class__.__name__}({self.body!r})"

    def __str__(self) -> str:
        return f"%({self.body})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.body == other.body


class ExpressionExpansion(ShellExpansion):
    """Node representing expression expansion, e.g. %[1+1]."""

    def __str__(self) -> str:
        return f"%[{self.body}]"


class MacroSubstitution(Node):
    """Node representing macro substitution, e.g. %version."""

    def __init__(self, body: str) -> None:
        _, self.prefix, self.name = re.split(r"([?!]*)", body, maxsplit=1)

    @formatted
    def __repr__(self) -> str:
        return f"MacroSubstitution({(self.prefix + self.name)!r})"

    def __str__(self) -> str:
        return f"%{self.prefix}{self.name}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.prefix == other.prefix and self.name == other.name


class EnclosedMacroSubstitution(Node):
    """Node representing macro substitution enclosed in brackets, e.g. %{?dist}."""

    def __init__(self, body: str) -> None:
        _, self.prefix, rest = re.split(r"([?!]*)", body, maxsplit=1)
        self.name: str
        self.args: List[str]
        self.name, *self.args = rest.split()

    @formatted
    def __repr__(self) -> str:
        args = (" " + " ".join(self.args)) if self.args else ""
        return f"EnclosedMacroSubstitution({(self.prefix + self.name + args)!r})"

    def __str__(self) -> str:
        args = (" " + " ".join(self.args)) if self.args else ""
        return f"%{{{self.prefix}{self.name}{args}}}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.prefix == other.prefix
            and self.name == other.name
            and self.args == other.args
        )


class ConditionalMacroExpansion(Node):
    """Node representing conditional macro expansion, e.g. %{?prerel:0.}."""

    def __init__(self, condition: str, body: List[Node]) -> None:
        _, self.prefix, self.name = re.split(r"([?!]*)", condition, maxsplit=1)
        self.body = body

    @formatted
    def __repr__(self) -> str:
        return (
            f"ConditionalMacroExpansion({(self.prefix + self.name)!r}, {self.body!r})"
        )

    def __str__(self) -> str:
        body = "".join(str(n) for n in self.body)
        return f"%{{{self.prefix}{self.name}:{body}}}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.prefix == other.prefix
            and self.name == other.name
            and self.body == other.body
        )


class BuiltinMacro(Node):
    """Node representing built-in macro, e.g. %{quote:Ancient Greek}."""

    def __init__(self, name: str, body: str) -> None:
        self.name = name
        self.body = body

    @formatted
    def __repr__(self) -> str:
        return f"BuiltinMacro({self.name!r}, {self.body!r})"

    def __str__(self) -> str:
        return f"%{{{self.name}:{self.body}}}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.name == other.name and self.body == other.body


class ValueParser:
    @classmethod
    def flatten(cls, nodes: List[Node]) -> Generator[Node, None, None]:
        """
        Generator that yields flattened nodes. Conditional macro expansions are treated
        as if their conditions were true and their bodies are flattened.

        Args:
            nodes: List of nodes to be flattened.

        Yields:
            Individual nodes.
        """
        for node in nodes:
            if isinstance(node, ConditionalMacroExpansion):
                yield from cls.flatten(node.body)
            else:
                yield node

    @classmethod
    def parse(cls, value: str) -> List[Node]:
        """
        Parses a value into a list of nodes.

        Follows the parsing logic of expandMacro() from rpm/rpmio/macro.c in RPM source.

        Args:
            value: Value string to parse.

        Returns:
            Parsed value as a list of nodes.

        Raises:
            UnterminatedMacroException if there is a macro that doesn't end.
        """
        pairs = {"(": ")", "{": "}", "[": "]"}

        def find_matching_parenthesis(index):
            level = 0
            for i in range(index, len(value)):
                if value[i] == "\\":
                    continue
                elif value[i] == pairs[value[index]]:
                    level -= 1
                    if level <= 0:
                        return i + 1
                elif value[i] == value[index]:
                    level += 1
            return None

        def find_macro_end(index):
            i = index
            if value[i] in pairs.keys():
                return find_matching_parenthesis(i)
            while i < len(value) and value[i] in "?!":
                i += 1
            if i < len(value) and value[i] == "-":
                i += 1
            while i < len(value) and (value[i].isalnum() or value[i] == "_"):
                i += 1
            if i + 1 < len(value) and value[i : i + 1] == "**":
                i += 2
            elif i < len(value) and value[i] in "*#":
                i += 1
            return i + 1

        result: List[Node] = []
        start = 0
        offset = 0
        while start < len(value):
            try:
                end = value.index("%", start + offset)
                try:
                    if value[end + 1] == "%":
                        offset = end + 2
                        continue
                except IndexError:
                    raise UnterminatedMacroException
            except ValueError:
                end = None
            if end is None:
                break
            if end > start:
                result.append(StringLiteral(value[start:end]))
            start = end
            end = find_macro_end(start + 1)
            if end is None:
                raise UnterminatedMacroException
            if value[start + 1] == "(":
                result.append(ShellExpansion(value[start + 2 : end - 1]))
            elif value[start + 1] == "[":
                result.append(ExpressionExpansion(value[start + 2 : end - 1]))
            elif value[start + 1] == "{":
                if ":" in value[start:end]:
                    condition, body = value[start + 2 : end - 1].split(":", maxsplit=1)
                    _, prefix, _ = re.split(r"([?!]*)", condition, maxsplit=1)
                    if "?" in prefix:
                        result.append(
                            ConditionalMacroExpansion(condition, cls.parse(body))
                        )
                    else:
                        result.append(BuiltinMacro(condition, body))
                else:
                    result.append(EnclosedMacroSubstitution(value[start + 2 : end - 1]))
            else:
                result.append(MacroSubstitution(value[start + 1 : end]))
            start = end
            offset = 0
        if value[start:]:
            result.append(StringLiteral(value[start:]))
        return result

    @classmethod
    def construct_regex(
        cls,
        value: str,
        modifiable_entities: Set[str],
        context: Optional["Specfile"] = None,
    ) -> Tuple[Pattern, Template]:
        """
        Parses the given value and constructs a regex that allows matching
        substrings of a different, but similar value to macro substitutions
        representing modifiable entities, and to modifiable substrings
        of the original value.
        Also constructs a corresponding template that allows updating
        the original value.

        For example, for nodes representing the string "1.%{version_minor}", assuming
        "version_minor" is a local macro definition (thus a modifiable entity),
        the resulting regex would be "^(?P<sub_0>)\\.(?P<version_minor>.+?)$",
        and the corresponding template would be "${sub_0}.%{version_minor}".
        If a requested new value would be a match to this regex, the "version_minor"
        macro definition could be modified with the matching substring and the final
        value could be determined by performing a substitution on the template
        with groupdict of the match.

        Args:
            value: Value string to parse.
            modifiable_entities: Names of modifiable entities, i.e. local macro definitions
              and tags.
            context: `Specfile` instance that defines the context for macro expansions.

        Returns:
            Tuple in the form of (constructed regex, corresponding template).
        """

        def expand(s):
            if context:
                result = context.expand(
                    s, skip_parsing=getattr(expand, "skip_parsing", False)
                )
                # parse only once
                expand.skip_parsing = True
                return result
            return Macros.expand(s)

        def flatten(nodes):
            # get rid of conditional macro expansions
            result = []
            for node in nodes:
                if isinstance(node, ConditionalMacroExpansion):
                    # evaluate the condition
                    if expand(f"%{node.prefix}{node.name}"):
                        result.append(f"%{{{node.prefix}{node.name}:")
                        result.extend(flatten(node.body))
                        result.append("}")
                    else:
                        result.append(str(node))
                else:
                    result.append(node)
            return result

        nodes = cls.parse(value)

        # convert nodes into constant, variable and group tokens
        tokens = []
        for node in flatten(nodes):
            if isinstance(node, str):
                tokens.append(("c", "", node))
            elif isinstance(node, StringLiteral):
                tokens.append(("v", node.value, ""))
            elif isinstance(node, (ShellExpansion, ExpressionExpansion, BuiltinMacro)):
                const = expand(str(node))
                tokens.append(("c", const, str(node)))
            elif isinstance(node, MacroSubstitution):
                if node.prefix.count("!") % 2 == 0 and node.name in modifiable_entities:
                    tokens.append(("g", node.name, str(node)))
                else:
                    const = expand(str(node))
                    tokens.append(("c", const, str(node)))
            elif isinstance(node, EnclosedMacroSubstitution):
                if (
                    node.prefix.count("!") % 2 == 0
                    and not node.args
                    and node.name in modifiable_entities
                ):
                    tokens.append(("g", node.name, str(node)))
                else:
                    const = expand(str(node))
                    tokens.append(("c", const, str(node)))

        def escape(s):
            return s.replace("$", "$$")

        # squash constants and variables
        _tokens = []
        for key, grp in itertools.groupby(tokens, lambda x: x[0]):
            if key in ("c", "v"):
                regex = template = ""
                for _, r, t in grp:
                    regex += r
                    template += t
                _tokens.append((key, regex, template))
            else:
                groups = list(grp)
                if len(groups) > 1:
                    # there are unseparated groups, reliable match is impossible
                    return re.compile("^$"), Template(escape(value))
                _tokens.extend(groups)
        tokens = _tokens

        def is_group_nearby(tokens, index, after):
            for token in tokens[index + 1 :] if after else reversed(tokens[:index]):
                if token[0] == "c":
                    # skip empty constants
                    if token[1]:
                        return False
                elif token[0] == "g":
                    return True
                elif token[0] == "v":
                    return False
            return False

        # construct regex and template
        regex = "^"
        template = ""
        named_groups = []
        sub = 0
        for i, token in enumerate(tokens):
            if token[0] == "c":
                regex += re.escape(token[1])
                template += escape(token[2])
            elif token[0] == "g":
                group = token[1]
                if group in named_groups:
                    # this group name already exists, make a backreference
                    regex += f"(?P={group})"
                else:
                    regex += f"(?P<{group}>.+?)"
                    named_groups.append(group)
                template += escape(token[2])
            elif token[0] == "v":
                value = token[1]
                # make sure there is at least one constant character between groups
                var_regex = "^"
                if is_group_nearby(tokens, i, False):
                    var_regex += r"(?P<prefix>\w*[^\w])"
                var_regex += "(?P<value>.*)"
                if is_group_nearby(tokens, i, True):
                    var_regex += r"(?P<suffix>[^\w]\w*)"
                var_regex += "$"
                m = re.match(var_regex, value)
                if not m:
                    regex += re.escape(value)
                    template += escape(value)
                    continue
                d = m.groupdict()
                prefix = d.get("prefix", "")
                value = d.get("value", "")
                suffix = d.get("suffix", "")
                regex += re.escape(prefix)
                template += escape(prefix)
                if value:
                    # make a substitution group
                    group = f"{SUBSTITUTION_GROUP_PREFIX}{sub}"
                    regex += f"(?P<{group}>.+?)"
                    template += f"${{{group}}}"
                    named_groups.append(group)
                    sub += 1
                regex += re.escape(suffix)
                template += escape(suffix)
        regex += "$"
        return re.compile(regex), Template(template)



#!/usr/bin/env python3

# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from pathlib import Path
from typing import List

import click
from pyparsing import (
    Combine,
    Group,
    Literal,
    ParserElement,
    Suppress,
    Word,
    c_style_comment,
    dbl_quoted_string,
    delimited_list,
    pyparsing_common,
    remove_quotes,
    srange,
)

# automatically suppress all string literals
ParserElement.inline_literals_using(Suppress)


def extract_sections(filename: Path) -> List[str]:
    """
    Extracts section names from a constant array looking like this:

    static const struct PartRec {
        int part;
        size_t len;
        const char * token;
    } partList[] = {
        { PART_PREAMBLE,      LEN_AND_STR("%package")},
        { PART_PREP,          LEN_AND_STR("%prep")},
        ...
        {0, 0, 0}
    };

    Args:
        filename: Path to the file to extract sections from.

    Returns:
        List of section names.
    """
    constant = Word(srange("[A-Z_]")).suppress()
    name = dbl_quoted_string.set_parse_action(remove_quotes)
    macro = "LEN_AND_STR(" + name + ")"
    item = "{" + Suppress(constant) + "," + macro + "}"
    sentinel = Suppress("{" + delimited_list(Literal("0")) + "}")
    parser = (
        Suppress("partList[]") + "=" + "{" + delimited_list(item) + "," + sentinel + "}"
    )
    parser.ignore(c_style_comment)
    result = parser.search_string(filename.read_text(), max_matches=1)
    if not result:
        return []
    return [s.lstrip("%") for s in result[0]]


def extract_tags(filename: Path, with_args: bool = False) -> List[str]:
    """
    Extracts tag names from a constant array looking like this:

    static struct PreambleRec_s const preambleList[] = {
        {RPMTAG_NAME,       0, 0, 1, LEN_AND_STR("name")},
        {RPMTAG_VERSION,        0, 0, 1, LEN_AND_STR("version")},
        ...
        {0, 0, 0, 0}
    };

    Args:
        filename: Path to the file to extract tags from.
        with_args: Include only tags that accept arguments.

    Returns:
        List of tag names.
    """
    constant = Word(srange("[A-Z_]")).suppress()
    name = dbl_quoted_string("name").set_parse_action(remove_quotes)
    macro = "LEN_AND_STR(" + name + ")"
    number = pyparsing_common.number
    item = Group(
        "{"
        + constant
        + ","
        + number("type")
        + ","
        + number.suppress()
        + ","
        + number.suppress()
        + ","
        + macro
        + "}"
    )
    sentinel = Suppress("{" + delimited_list(Literal("0")) + "}")
    parser = (
        Suppress("preambleList[]")
        + "="
        + "{"
        + delimited_list(item)
        + ","
        + sentinel
        + "}"
    )
    parser.ignore(c_style_comment)
    result = parser.search_string(filename.read_text(), max_matches=1)
    if not result:
        return []
    return [t.name for t in result[0] if not with_args or t.type > 0]


def extract_arches(filename: Path) -> List[str]:
    """
    Extracts arch names from a list looking like this:

    arch_canon: athlon: athlon  1
    arch_canon: geode:  geode   1
    ...
    arch_canon: IP: sgi 7
    ...
    arch_canon: atariclone: m68kmint    13
    ...

    Args:
        filename: Path to the file to extract arches from.

    Returns:
        List of arch names.
    """
    identifier = pyparsing_common.identifier
    number = pyparsing_common.number
    parser = (
        Suppress("arch_canon:")
        + Combine(identifier + ":")
        + identifier.suppress()
        + number.suppress()
    )
    result = parser.search_string(filename.read_text())
    return [a[0] for a in result]


@click.group()
def extract():
    pass


@extract.command(
    help="Extract section names from the specified source file (typically build/parseSpec.c)."
)
@click.argument(
    "filename",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    required=True,
)
def sections(filename: Path) -> None:
    for section in extract_sections(filename):
        click.echo(section)


@extract.command(
    help="Extract tag names from the specified source file (typically build/parsePreamble.c)."
)
@click.argument(
    "filename",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    required=True,
)
@click.option(
    "--with-args",
    is_flag=True,
    default=False,
    help="List only tags that accept arguments.",
)
def tags(filename: Path, with_args: bool = False) -> None:
    for tag in extract_tags(filename, with_args):
        click.echo(tag)


@extract.command(
    help="Extract arch names from the specified template file (typically rpmrc.in)."
)
@click.argument(
    "filename",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    required=True,
)
def arches(filename: Path) -> None:
    for arch in extract_arches(filename):
        click.echo(arch)


if __name__ == "__main__":
    extract()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import copy
import datetime

import pytest

from specfile.changelog import Changelog, ChangelogEntry
from specfile.sections import Section


@pytest.mark.parametrize(
    "header, evr",
    [
        ("* Thu Jan 04 2007 Michael Schwendt <mschwendt@fedoraproject.org>", None),
        (
            "* Mon Jul 13 2020 Tom Stellard <tstellar@redhat.com> 4.0-0.4.pre2",
            "4.0-0.4.pre2",
        ),
        ("* Fri Jul 20 2018 Gwyn Ciesla <limburgher@gmail.com> - 0.52-6", "0.52-6"),
        (
            "* Mon Feb 23 2009 Fedora Release Engineering <rel-eng@lists.fedoraproject.org> "
            "- 1.23-3.20081106gitbe42b4",
            "1.23-3.20081106gitbe42b4",
        ),
        (
            "* Thu Feb 04 2016 Marcin Zajaczkowski <mszpak ATT wp DOTT pl> - 1:0.9.10-6",
            "1:0.9.10-6",
        ),
        (
            "* Mon Jan 03 2022 Fedora Kernel Team <kernel-team@fedoraproject.org> [5.16-0.rc8.55]",
            "5.16-0.rc8.55",
        ),
        ("* Wed Jan 23 2002 Karsten Hopp <karsten@redhat.de> (4.6-1)", "4.6-1"),
        (
            "* Thu Apr  9 2015 Jeffrey C. Ollie <jeff@ocjtech.us> - 13.3.2-1:",
            "13.3.2-1",
        ),
    ],
)
def test_entry_evr(header, evr):
    assert ChangelogEntry(header, [""]).evr == evr


@pytest.mark.parametrize(
    "header, extended",
    [
        ("* Tue May 4 2021 Nikola Forró <nforro@redhat.com> - 0.1-1", False),
        ("* Tue May  4 2021 Nikola Forró <nforro@redhat.com> - 0.1-1", False),
        (
            "* Thu Jul 22 2021 Fedora Release Engineering <releng@fedoraproject.org> - 0.1-2",
            False,
        ),
        (
            "* Mon Oct 18 12:34:45 CEST 2021 Nikola Forró <nforro@redhat.com> - 0.2-1",
            True,
        ),
    ],
)
def test_entry_has_extended_timestamp(header, extended):
    assert ChangelogEntry(header, [""]).extended_timestamp == extended


@pytest.mark.parametrize(
    "header, padding",
    [
        ("* Tue May 4 2021 Nikola Forró <nforro@redhat.com> - 0.1-1", ""),
        ("* Tue May 04 2021 Nikola Forró <nforro@redhat.com> - 0.1-1", "0"),
        ("* Tue May  4 2021 Nikola Forró <nforro@redhat.com> - 0.1-1", " "),
        (
            "* Thu Jul 22 2021 Fedora Release Engineering <releng@fedoraproject.org> - 0.1-2",
            "",
        ),
        (
            "* Mon Oct  18 12:34:45 CEST 2021 Nikola Forró <nforro@redhat.com> - 0.2-1",
            " ",
        ),
    ],
)
def test_entry_day_of_month_padding(header, padding):
    assert ChangelogEntry(header, [""]).day_of_month_padding == padding


@pytest.mark.parametrize(
    "since, until, evrs",
    [
        (None, None, ["0.1-1", "0.1-2", "0.2-1", "0.2-2"]),
        ("0.1-1", None, ["0.1-1", "0.1-2", "0.2-1", "0.2-2"]),
        ("0.1-2", None, ["0.1-2", "0.2-1", "0.2-2"]),
        ("0.2-1", None, ["0.2-1", "0.2-2"]),
        (None, "0.2-2", ["0.1-1", "0.1-2", "0.2-1", "0.2-2"]),
        (None, "0.2-1", ["0.1-1", "0.1-2", "0.2-1"]),
        (None, "0.1-2", ["0.1-1", "0.1-2"]),
        ("0.1-1", "0.2-2", ["0.1-1", "0.1-2", "0.2-1", "0.2-2"]),
        ("0.1-2", "0.2-1", ["0.1-2", "0.2-1"]),
        ("0.2-1", "0.2-1", ["0.2-1"]),
        ("0.2-2", "0.1-1", []),
        ("0.0.1-1", None, ["0.1-1", "0.1-2", "0.2-1", "0.2-2"]),
        ("0.3-1", None, []),
        (None, "0.0.1-1", []),
        (None, "0.3-1", ["0.1-1", "0.1-2", "0.2-1", "0.2-2"]),
        ("0.0.1-1", "0.3-1", ["0.1-1", "0.1-2", "0.2-1", "0.2-2"]),
    ],
)
def test_filter(since, until, evrs):
    changelog = Changelog(
        [
            ChangelogEntry.assemble(
                datetime.date(2021, 5, 4),
                "Nikola Forró <nforro@redhat.com>",
                ["- first version", "  resolves: #999999999"],
                "0.1-1",
                append_newline=False,
            ),
            ChangelogEntry.assemble(
                datetime.date(2021, 7, 22),
                "Fedora Release Engineering <releng@fedoraproject.org>",
                ["- Rebuilt for https://fedoraproject.org/wiki/Fedora_35_Mass_Rebuild"],
                "0.1-2",
            ),
            ChangelogEntry.assemble(
                datetime.datetime(2021, 10, 18, 12, 34, 45),
                "Nikola Forró <nforro@redhat.com>",
                ["- new upstream release"],
                "0.2-1",
            ),
            ChangelogEntry.assemble(
                datetime.datetime(2022, 1, 13, 8, 12, 41),
                "Nikola Forró <nforro@redhat.com>",
                ["- rebuilt"],
                "0.2-2",
            ),
        ]
    )
    assert [e.evr for e in changelog.filter(since=since, until=until)] == evrs


def test_parse():
    changelog = Changelog.parse(
        Section(
            "changelog",
            data=[
                "* Fri Jan 27 2023 Nikola Forró <nforro@redhat.com> - 0.4-1",
                "",
                "* this is also a valid entry",
                "",
                "* Fri Jan 27 2023 Nikola Forró <nforro@redhat.com> - 0.3-2",
                "* this is a valid entry",
                "",
                "* Mon Nov 21 2022 Nikola Forró <nforro@redhat.com> - 0.3-1",
                "- this is a formatted",
                "  changelog entry",
                "",
                "- here is another item",
                "",
                "* Thu Jan 13 08:12:41 UTC 2022 Nikola Forró <nforro@redhat.com> - 0.2-2",
                "- rebuilt",
                "",
                "* Mon Oct 18 12:34:45 CEST 2021 Nikola Forró <nforro@redhat.com> - 0.2-1",
                "- new upstream release",
                "",
                "* Thu Jul 22 2021 Fedora Release Engineering <releng@fedoraproject.org> - 0.1-2",
                "- Rebuilt for https://fedoraproject.org/wiki/Fedora_35_Mass_Rebuild",
                "",
                "* Tue May 04 2021 Nikola Forró <nforro@redhat.com> - 0.1-1",
                "- first version",
                "  resolves: #999999999",
            ],
        )
    )
    assert len(changelog) == 7
    assert (
        changelog[0].header
        == "* Tue May 04 2021 Nikola Forró <nforro@redhat.com> - 0.1-1"
    )
    assert changelog[0].content == ["- first version", "  resolves: #999999999"]
    assert not changelog[0].extended_timestamp
    assert (
        changelog[1].header
        == "* Thu Jul 22 2021 Fedora Release Engineering <releng@fedoraproject.org> - 0.1-2"
    )
    assert changelog[1].content == [
        "- Rebuilt for https://fedoraproject.org/wiki/Fedora_35_Mass_Rebuild"
    ]
    assert not changelog[1].extended_timestamp
    assert (
        changelog[2].header
        == "* Mon Oct 18 12:34:45 CEST 2021 Nikola Forró <nforro@redhat.com> - 0.2-1"
    )
    assert changelog[2].content == ["- new upstream release"]
    assert changelog[2].extended_timestamp
    assert (
        changelog[3].header
        == "* Thu Jan 13 08:12:41 UTC 2022 Nikola Forró <nforro@redhat.com> - 0.2-2"
    )
    assert changelog[3].content == ["- rebuilt"]
    assert changelog[3].extended_timestamp
    assert (
        changelog[4].header
        == "* Mon Nov 21 2022 Nikola Forró <nforro@redhat.com> - 0.3-1"
    )
    assert changelog[4].content == [
        "- this is a formatted",
        "  changelog entry",
        "",
        "- here is another item",
    ]
    assert not changelog[4].extended_timestamp
    assert (
        changelog[5].header
        == "* Fri Jan 27 2023 Nikola Forró <nforro@redhat.com> - 0.3-2"
    )
    assert changelog[5].content == [
        "* this is a valid entry",
    ]
    assert not changelog[5].extended_timestamp
    assert (
        changelog[6].header
        == "* Fri Jan 27 2023 Nikola Forró <nforro@redhat.com> - 0.4-1"
    )
    assert changelog[6].content == [
        "",
        "* this is also a valid entry",
    ]
    assert not changelog[6].extended_timestamp


def test_get_raw_section_data():
    tzinfo = datetime.timezone(datetime.timedelta(hours=2), name="CEST")
    changelog = Changelog(
        [
            ChangelogEntry.assemble(
                datetime.date(2021, 5, 4),
                "Nikola Forró <nforro@redhat.com>",
                ["- first version", "  resolves: #999999999"],
                "0.1-1",
                append_newline=False,
            ),
            ChangelogEntry.assemble(
                datetime.date(2021, 7, 22),
                "Fedora Release Engineering <releng@fedoraproject.org>",
                ["- Rebuilt for https://fedoraproject.org/wiki/Fedora_35_Mass_Rebuild"],
                "0.1-2",
            ),
            ChangelogEntry.assemble(
                datetime.datetime(2021, 10, 18, 12, 34, 45, tzinfo=tzinfo),
                "Nikola Forró <nforro@redhat.com>",
                ["- new upstream release"],
                "0.2-1",
            ),
            ChangelogEntry.assemble(
                datetime.datetime(2022, 1, 13, 8, 12, 41),
                "Nikola Forró <nforro@redhat.com>",
                ["- rebuilt"],
                "0.2-2",
            ),
            ChangelogEntry.assemble(
                datetime.date(2022, 11, 21),
                "Nikola Forró <nforro@redhat.com>",
                [
                    "- this is a formatted",
                    "  changelog entry",
                    "",
                    "- here is another item",
                ],
                "0.3-1",
            ),
            ChangelogEntry.assemble(
                datetime.date(2023, 1, 27),
                "Nikola Forró <nforro@redhat.com>",
                [
                    "* this is a valid entry",
                ],
                "0.3-2",
            ),
            ChangelogEntry.assemble(
                datetime.date(2023, 1, 27),
                "Nikola Forró <nforro@redhat.com>",
                [
                    "",
                    "* this is also a valid entry",
                ],
                "0.4-1",
            ),
        ]
    )
    assert changelog.get_raw_section_data() == [
        "* Fri Jan 27 2023 Nikola Forró <nforro@redhat.com> - 0.4-1",
        "",
        "* this is also a valid entry",
        "",
        "* Fri Jan 27 2023 Nikola Forró <nforro@redhat.com> - 0.3-2",
        "* this is a valid entry",
        "",
        "* Mon Nov 21 2022 Nikola Forró <nforro@redhat.com> - 0.3-1",
        "- this is a formatted",
        "  changelog entry",
        "",
        "- here is another item",
        "",
        "* Thu Jan 13 08:12:41 UTC 2022 Nikola Forró <nforro@redhat.com> - 0.2-2",
        "- rebuilt",
        "",
        "* Mon Oct 18 12:34:45 CEST 2021 Nikola Forró <nforro@redhat.com> - 0.2-1",
        "- new upstream release",
        "",
        "* Thu Jul 22 2021 Fedora Release Engineering <releng@fedoraproject.org> - 0.1-2",
        "- Rebuilt for https://fedoraproject.org/wiki/Fedora_35_Mass_Rebuild",
        "",
        "* Tue May 04 2021 Nikola Forró <nforro@redhat.com> - 0.1-1",
        "- first version",
        "  resolves: #999999999",
    ]


def test_copy_changelog():
    changelog = Changelog(
        [
            ChangelogEntry.assemble(
                datetime.date(2021, 5, 4),
                "Nikola Forró <nforro@redhat.com>",
                ["- first version", "  resolves: #999999999"],
                "0.1-1",
                append_newline=False,
            ),
        ]
    )
    shallow_copy = copy.copy(changelog)
    assert shallow_copy == changelog
    assert shallow_copy is not changelog
    assert shallow_copy[0] is changelog[0]
    deep_copy = copy.deepcopy(changelog)
    assert deep_copy == changelog
    assert deep_copy is not changelog
    assert deep_copy[0] is not changelog[0]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

# valid section names as defined in build/parseSpec.c in RPM source
SECTION_NAMES = {
    "build",
    "changelog",
    "check",
    "clean",
    "conf",
    "description",
    "end",
    "files",
    "filetrigger",
    "filetriggerin",
    "filetriggerpostun",
    "filetriggerun",
    "generate_buildrequires",
    "install",
    "package",
    "patchlist",
    "post",
    "posttrans",
    "postun",
    "postuntrans",
    "pre",
    "prep",
    "pretrans",
    "preun",
    "preuntrans",
    "sepolicy",
    "sourcelist",
    "transfiletrigger",
    "transfiletriggerin",
    "transfiletriggerpostun",
    "transfiletriggerun",
    "trigger",
    "triggerin",
    "triggerpostun",
    "triggerprein",
    "triggerun",
    "verifyscript",
}

SIMPLE_SCRIPT_SECTIONS = {
    "build",
    "check",
    "clean",
    "conf",
    "generate_buildrequires",
    "install",
}

SCRIPT_SECTIONS = {
    "filetrigger",
    "filetriggerin",
    "filetriggerpostun",
    "filetriggerun",
    "post",
    "posttrans",
    "postun",
    "postuntrans",
    "pre",
    "pretrans",
    "preun",
    "preuntrans",
    "transfiletrigger",
    "transfiletriggerin",
    "transfiletriggerpostun",
    "transfiletriggerun",
    "trigger",
    "triggerin",
    "triggerpostun",
    "triggerprein",
    "triggerun",
    "verifyscript",
}

# valid section option strings taken from build/parseSpec.c, build/parsePreamble.c,
# build/parseDescription.c, build/parseFiles.c, build/parsePolicies.c and build/parseScript.c
# in RPM source
SECTION_OPTIONS = {
    "description": "n:l:",
    "files": "n:f:",
    "package": "n:",
    "sepolicy": "n:",
}
SECTION_OPTIONS.update({s: "n:f:p:P:eq" for s in SCRIPT_SECTIONS})

# valid tag names as defined in build/parsePreamble.c in RPM source
TAG_NAMES = {
    "autoprov",
    "autoreq",
    "autoreqprov",
    "bugurl",
    "buildarch",
    "buildarchitectures",
    "buildconflicts",
    "buildprereq",
    "buildrequires",
    "buildroot",
    "conflicts",
    "distribution",
    "disttag",
    "disturl",
    "docdir",
    "enhances",
    "epoch",
    "excludearch",
    "excludeos",
    "exclusivearch",
    "exclusiveos",
    "group",
    "icon",
    "license",
    "modularitylabel",
    "name",
    "nopatch",
    "nosource",
    "obsoletes",
    "orderwithrequires",
    "packager",
    "patch",
    "prefix",
    "prefixes",
    "prereq",
    "provides",
    "recommends",
    "release",
    "removepathpostfixes",
    "requires",
    "source",
    "sourcelicense",
    "suggests",
    "summary",
    "supplements",
    "translationurl",
    "upstreamreleases",
    "url",
    "vcs",
    "vendor",
    "version",
}

# tags that can optionally have an argument (language or qualifier)
TAGS_WITH_ARG = {
    "group",
    "orderwithrequires",
    "prereq",
    "requires",
    "summary",
}

# canonical architecture names as defined in rpmrc.in in RPM source
ARCH_NAMES = {
    "IP",
    "aarch64",
    "alpha",
    "alphaev5",
    "alphaev56",
    "alphaev6",
    "alphaev67",
    "alphapca56",
    "amd64",
    "armv3l",
    "armv4b",
    "armv4l",
    "armv5tejl",
    "armv5tel",
    "armv5tl",
    "armv6hl",
    "armv6l",
    "armv7hl",
    "armv7hnl",
    "armv7l",
    "armv8hl",
    "armv8l",
    "atariclone",
    "atarist",
    "atariste",
    "ataritt",
    "athlon",
    "em64t",
    "falcon",
    "geode",
    "hades",
    "i370",
    "i386",
    "i486",
    "i586",
    "i686",
    "ia32e",
    "ia64",
    "loongarch64",
    "m68k",
    "m68kmint",
    "milan",
    "mips",
    "mips64",
    "mips64el",
    "mips64r6",
    "mips64r6el",
    "mipsel",
    "mipsr6",
    "mipsr6el",
    "pentium3",
    "pentium4",
    "ppc",
    "ppc32dy4",
    "ppc64",
    "ppc64iseries",
    "ppc64le",
    "ppc64p7",
    "ppc64pseries",
    "ppc8260",
    "ppc8560",
    "ppciseries",
    "ppcpseries",
    "riscv",
    "riscv64",
    "rs6000",
    "s390",
    "s390x",
    "sh",
    "sh3",
    "sh4",
    "sh4a",
    "sparc",
    "sparc64",
    "sparc64v",
    "sparcv8",
    "sparcv9",
    "sparcv9v",
    "sun4",
    "sun4c",
    "sun4d",
    "sun4m",
    "sun4u",
    "x86_64",
    "x86_64_v2",
    "x86_64_v3",
    "x86_64_v4",
    "xtensa",
}



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import collections
import copy
import re
from abc import ABC
from typing import Any, Dict, List, Optional, SupportsIndex, Union, cast, overload

from specfile.formatter import formatted
from specfile.options import Options
from specfile.sections import Section
from specfile.utils import split_conditional_macro_expansion


def valid_prep_macro(name: str) -> bool:
    return name in ("setup", "autosetup", "autopatch") or name.startswith("patch")


class PrepMacro(ABC):
    """
    Class that represents a %prep macro.

    Attributes:
        name: Literal name of the macro.
        options: Options of the macro.
    """

    CANONICAL_NAME: str
    OPTSTRING: str
    DEFAULTS: Dict[str, Union[bool, int, str]]

    def __init__(
        self,
        name: str,
        options: Options,
        delimiter: str,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
        preceding_lines: Optional[List[str]] = None,
    ) -> None:
        """
        Constructs a `PrepMacro` object.

        Args:
            name: Literal name of the macro.
            options: Options of the macro.
            delimiter: Delimiter separating name and option string.
            prefix: Characters preceding the macro on a line.
            suffix: Characters following the macro on a line.
            preceding_lines: Lines of the %prep section preceding the macro.

        Returns:
            Constructed instance of `PrepMacro` class.
        """
        self.name = name
        self.options = copy.deepcopy(options)
        self._delimiter = delimiter
        self._prefix = prefix or ""
        self._suffix = suffix or ""
        self._preceding_lines = (
            preceding_lines.copy() if preceding_lines is not None else []
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PrepMacro):
            return NotImplemented
        return (
            self.name == other.name
            and self.options == other.options
            and self._delimiter == other._delimiter
            and self._prefix == other._prefix
            and self._suffix == other._suffix
            and self._preceding_lines == other._preceding_lines
        )

    @formatted
    def __repr__(self) -> str:
        # determine class name dynamically so that inherited classes
        # don't have to reimplement __repr__()
        return (
            f"{self.__class__.__name__}({self.name!r}, {self.options!r}, "
            f"{self._delimiter!r}, {self._prefix!r}, {self._suffix!r}, "
            f"{self._preceding_lines!r})"
        )

    def get_raw_data(self) -> List[str]:
        options = str(self.options)
        # ensure delimiter is not empty when there are any options
        if options and not self._delimiter:
            self._delimiter = " "
        return self._preceding_lines + [
            f"{self._prefix}{self.name}{self._delimiter}{options}{self._suffix}"
        ]


class SetupMacro(PrepMacro):
    """Class that represents a %setup macro."""

    CANONICAL_NAME: str = "%setup"
    OPTSTRING: str = "a:b:cDn:Tq"
    DEFAULTS: Dict[str, Union[bool, int, str]] = {
        "n": "%{name}-%{version}",
    }


class PatchMacro(PrepMacro):
    """Class that represents a %patch macro."""

    CANONICAL_NAME: str = "%patch"
    OPTSTRING: str = "P:p:REb:z:F:d:o:Z"
    DEFAULTS: Dict[str, Union[bool, int, str]] = {}

    @property
    def number(self) -> int:
        """Number of the %patch macro."""
        if self.options.P is not None:
            return int(self.options.P)
        tokens = re.split(r"(\d+)", self.name, maxsplit=1)
        if len(tokens) > 1:
            return int(tokens[1])
        return 0

    @number.setter
    def number(self, value: int) -> None:
        if self.options.P is not None:
            self.options.P = value
            return
        self.name = f"{self.CANONICAL_NAME}{value}"


class AutosetupMacro(PrepMacro):
    """Class that represents an %autosetup macro."""

    CANONICAL_NAME: str = "%autosetup"
    OPTSTRING: str = "a:b:cDn:TvNS:p:"
    DEFAULTS: Dict[str, Union[bool, int, str]] = {
        "n": "%{name}-%{version}",
        "S": "patch",
    }


class AutopatchMacro(PrepMacro):
    """Class that represents an %autopatch macro."""

    CANONICAL_NAME: str = "%autopatch"
    OPTSTRING: str = "vp:m:M:"
    DEFAULTS: Dict[str, Union[bool, int, str]] = {}


class PrepMacros(collections.UserList):
    """
    Class that represents a list of %prep macros.

    Attributes:
        data: List of individual %prep macros.
    """

    def __init__(
        self,
        data: Optional[List[PrepMacro]] = None,
        remainder: Optional[List[str]] = None,
    ) -> None:
        """
        Constructs a `PrepMacros` object.

        Args:
            data: List of individual %prep macros.
            remainder: Leftover lines in the section.

        Returns:
            Constructed instance of `PrepMacros` class.
        """
        super().__init__()
        if data is not None:
            self.data = data.copy()
        self._remainder = remainder.copy() if remainder is not None else []

    @formatted
    def __repr__(self) -> str:
        return f"PrepMacros({self.data!r}, {self._remainder!r})"

    def __contains__(self, item: object) -> bool:
        if isinstance(item, type):
            return any(isinstance(m, item) for m in self.data)
        return any(
            m.name.startswith(item) if item == "%patch" else m.name == item
            for m in self.data
        )

    @overload
    def __getitem__(self, i: SupportsIndex) -> PrepMacro:
        pass

    @overload
    def __getitem__(self, i: slice) -> "PrepMacros":
        pass

    def __getitem__(self, i):
        if isinstance(i, slice):
            return PrepMacros(self.data[i], self._remainder)
        else:
            return self.data[i]

    def __delitem__(self, i: Union[SupportsIndex, slice]) -> None:
        def delete(index):
            preceding_lines = self.data[index]._preceding_lines.copy()
            del self.data[index]
            if index < len(self.data):
                self.data[index]._preceding_lines = (
                    preceding_lines + self.data[index]._preceding_lines
                )
            else:
                self._remainder = preceding_lines + self._remainder

        if isinstance(i, slice):
            for index in reversed(range(len(self.data))[i]):
                delete(index)
        else:
            delete(i)

    def __getattr__(self, name: str) -> PrepMacro:
        if not valid_prep_macro(name):
            return super().__getattribute__(name)
        try:
            return self.data[self.find(f"%{name}")]
        except ValueError:
            raise AttributeError(name)

    def __delattr__(self, name: str) -> None:
        if not valid_prep_macro(name):
            return super().__delattr__(name)
        try:
            self.__delitem__(self.find(f"%{name}"))
        except ValueError:
            raise AttributeError(name)

    def copy(self) -> "PrepMacros":
        return copy.copy(self)

    def find(self, name: str) -> int:
        for i, macro in enumerate(self.data):
            if macro.name == name:
                return i
        raise ValueError

    def get_raw_data(self) -> List[str]:
        result = []
        for macro in self.data:
            result.extend(macro.get_raw_data())
        result.extend(self._remainder)
        return result


class Prep(collections.abc.Container):
    """
    Class that represents a %prep section.

    Attributes:
        macros: List of individual %prep macros.
    """

    def __init__(self, macros: PrepMacros) -> None:
        self.macros = macros.copy()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Prep):
            return NotImplemented
        return self.macros == other.macros

    @formatted
    def __repr__(self) -> str:
        return f"Prep({self.macros!r})"

    def __contains__(self, item: object) -> bool:
        return item in self.macros

    def __getattr__(self, name: str) -> PrepMacro:
        if not valid_prep_macro(name):
            return super().__getattribute__(name)
        return getattr(self.macros, name)

    def __delattr__(self, name: str) -> None:
        if not valid_prep_macro(name):
            return super().__delattr__(name)
        return delattr(self.macros, name)

    def add_patch_macro(self, number: int, **kwargs: Any) -> None:
        """
        Adds a new %patch macro with given number and options.

        Args:
            number: Macro number.
            P: The -P option (patch number).
            p: The -p option (strip number).
            R: The -R option (reverse).
            E: The -E option (remove empty files).
            b: The -b option (backup).
            z: The -z option (suffix).
            F: The -F option (fuzz factor).
            d: The -d option (working directory).
            o: The -o option (output file).
            Z: The -Z option (set UTC times).
        """
        options = Options([], PatchMacro.OPTSTRING, PatchMacro.DEFAULTS)
        for k, v in kwargs.items():
            setattr(options, k, v)
        macro = PatchMacro(PatchMacro.CANONICAL_NAME, options, " ")
        macro.number = number
        index, _ = min(
            ((i, m) for i, m in enumerate(self.macros) if isinstance(m, PatchMacro)),
            key=lambda im: abs(im[1].number - number),
            default=(len(self.macros), None),
        )
        if (
            index < len(self.macros)
            and cast(PatchMacro, self.macros[index]).number <= number
        ):
            index += 1
        self.macros.insert(index, macro)

    def remove_patch_macro(self, number: int) -> None:
        """
        Removes a %patch macro with given number.

        Args:
            number: Macro number.
        """
        index = next(
            (
                i
                for i, m in enumerate(self.macros)
                if isinstance(m, PatchMacro) and m.number == number
            ),
            None,
        )
        if index:
            del self.macros[index]

    @classmethod
    def parse(cls, section: Section) -> "Prep":
        """
        Parses a section into a `Prep` object.

        Args:
            section: %prep section.

        Returns:
            Constructed instance of `Prep` class.
        """
        macro_regex = re.compile(
            r"(?P<m>%(setup|patch\d*|autopatch|autosetup))(?P<d>\s*)(?P<o>.*?)$"
        )
        data = []
        buffer: List[str] = []
        for line in section:
            line, prefix, suffix = split_conditional_macro_expansion(line)
            m = macro_regex.search(line)
            if m:
                name, delimiter, option_string = (
                    m.group("m"),
                    m.group("d"),
                    m.group("o"),
                )
                prefix += line[: m.start("m")]
                suffix = line[m.end("o") :] + suffix
                klass = next(
                    (
                        klass
                        for klass in PrepMacro.__subclasses__()
                        if name.startswith(klass.CANONICAL_NAME)
                    ),
                    None,
                )
                if not klass:
                    buffer.append(line)
                    continue
                options = Options(
                    Options.tokenize(option_string),
                    klass.OPTSTRING,
                    klass.DEFAULTS,
                )
                data.append(klass(name, options, delimiter, prefix, suffix, buffer))
                buffer = []
            else:
                buffer.append(line)
        return cls(PrepMacros(data, buffer))

    def get_raw_section_data(self) -> List[str]:
        """
        Reconstructs section data from `Prep` object.

        Returns:
            List of lines forming the reconstructed section data.
        """
        return self.macros.get_raw_data()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import contextlib
import copy
import logging
import os
import re
import tempfile
from pathlib import Path
from typing import Any, Dict, Generator, List, Optional, Set, Tuple

import rpm

from specfile.context_management import capture_stderr
from specfile.exceptions import RPMException
from specfile.formatter import formatted
from specfile.macros import Macros
from specfile.sections import Section
from specfile.tags import Tags
from specfile.utils import get_filename_from_location
from specfile.value_parser import BuiltinMacro, ShellExpansion, ValueParser

logger = logging.getLogger(__name__)


class SpecParser:
    """
    Class that represents a spec file parser.

    Attributes:
        sourcedir: Path to sources and patches.
        macros: List of extra macro definitions.
        force_parse: Whether to attempt to parse the spec file even if one or more
          sources required to be present at parsing time are not available.
          Such sources include sources referenced from shell expansions
          in tag values and sources included using the %include directive.
        spec: `rpm.spec` instance representing parsed spec file.
        tainted: Indication that parsing of the spec file was forced and one or more
          sources required to be present at parsing time were not available
          and were replaced with dummy files.
    """

    def __init__(
        self,
        sourcedir: Path,
        macros: Optional[List[Tuple[str, str]]] = None,
        force_parse: bool = False,
    ) -> None:
        self.sourcedir = sourcedir
        self.macros = macros.copy() if macros is not None else []
        self.force_parse = force_parse
        self.spec = None
        self.tainted = False

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SpecParser):
            return NotImplemented
        return (
            self.sourcedir == other.sourcedir
            and self.macros == other.macros
            and self.force_parse == other.force_parse
        )

    @formatted
    def __repr__(self) -> str:
        return f"SpecParser({self.sourcedir!r}, {self.macros!r}, {self.force_parse!r})"

    def __deepcopy__(self, memo: Dict[int, Any]) -> "SpecParser":
        result = self.__class__.__new__(self.__class__)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k in ["spec", "tainted"]:
                continue
            setattr(result, k, copy.deepcopy(v, memo))
        result.spec = None
        result.tainted = False
        return result

    @contextlib.contextmanager
    def _make_dummy_sources(
        self, sources: Set[str], non_empty_sources: Set[str]
    ) -> Generator[List[Path], None, None]:
        """
        Context manager for creating temporary dummy sources to enable a spec file
        to be fully parsed by RPM.

        Args:
            sources: Set of sources to be faked with effectively empty files if they don't exist.
              File signatures will be generated according to extensions.
            non_empty_sources: Set of sources to be faked with text files containing
              dummy content if they don't exist.

        Yields:
            List of paths to each created dummy source.
        """
        # based on rpmFileIsCompressed() in rpmio/rpmfileutil.c in RPM source
        SIGNATURES = [
            (".bz2", b"BZh"),
            (".zip", b"PK00"),
            (".xz", b"\xfd\x37\x7a\x58\x5a\x00"),
            (".zst", b"\x28\xb5\x2f"),
            (".lz", b"LZIP"),
            (".lrz", b"LRZI"),
            (".gz", b"\x1f\x8b"),
            (".7z", b"7z\xbc\xaf\x27\x1c"),
        ]
        # number of bytes that RPM reads to determine the file type
        MAGIC_LENGTH = 13
        dummy_sources = []
        for source in sources:
            filename = get_filename_from_location(source)
            if not filename:
                continue
            path = self.sourcedir / filename
            if path.is_file():
                continue
            dummy_sources.append(path)
            for ext, magic in SIGNATURES:
                if filename.endswith(ext):
                    path.write_bytes(magic.ljust(MAGIC_LENGTH, b"\x00"))
                    break
            else:
                path.write_bytes(MAGIC_LENGTH * b"\x00")
        for source in non_empty_sources:
            filename = get_filename_from_location(source)
            if not filename:
                continue
            path = self.sourcedir / filename
            if path.is_file():
                continue
            dummy_sources.append(path)
            path.write_text("DUMMY")
        try:
            yield dummy_sources
        finally:
            for path in dummy_sources:
                path.unlink()

    @contextlib.contextmanager
    def _sanitize_environment(self) -> Generator[os._Environ, None, None]:
        """
        Context manager for sanitizing the environment for shell expansions.

        Temporarily sets LANG and LC_ALL to C.UTF-8 locale.

        Yields:
            Sanitized environment.
        """
        env = os.environ.copy()

        def restore(key):
            try:
                os.environ[key] = env[key]
            except KeyError:
                del os.environ[key]

        keys = ["LANG", "LC_ALL"]
        for key in keys:
            os.environ[key] = "C.UTF-8"
        try:
            yield os.environ
        finally:
            for key in keys:
                restore(key)

    def _do_parse(
        self, content: str, extra_macros: Optional[List[Tuple[str, str]]] = None
    ) -> Tuple[rpm.spec, bool]:
        """
        Parses the content of a spec file.

        Args:
            content: String representing the content of a spec file.
            extra_macros: List of extra macro definitions.

        Returns:
            Tuple of (parsed spec file as `rpm.spec` instance, indication whether
              at least one file to be included was ignored).

        Raises:
            RPMException, if parsing error occurs.
        """

        def get_rpm_spec(content, flags):
            Macros.reinit()
            for name, value in self.macros + (extra_macros or []):
                Macros.define(name, value)
            Macros.define("_sourcedir", str(self.sourcedir))
            with tempfile.NamedTemporaryFile() as tmp:
                tmp.write(content.encode())
                tmp.flush()
                try:
                    with self._sanitize_environment():
                        with capture_stderr() as stderr:
                            return rpm.spec(tmp.name, flags)
                except ValueError as e:
                    raise RPMException(stderr=stderr) from e

        def collect_sources_referenced_from_tags(content):
            # collect sources referenced from shell expansions in tag values
            sources = set()
            # source references: %SOURCEN, %{SOURCEN}, %{S:N}
            source_ref_regex = re.compile(r"%((?P<b>{)?[?!]*SOURCE\d+(?(b)})|{S:\d+})")
            for tag in Tags.parse(Section("package", data=content.splitlines())):
                # we can expand macros here because the first non-build parse,
                # even though it failed, populated the macro context
                if Macros.expand(tag.value):
                    # tag value doesn't expand to an empty string, so it won't
                    # break parsing, we can skip this tag
                    continue
                refs = []
                for node in ValueParser.flatten(ValueParser.parse(tag.value)):
                    if isinstance(node, ShellExpansion):
                        for m in source_ref_regex.finditer(node.body):
                            refs.append(m.group(0))
                for ref in refs:
                    # we can expand macros here because the first non-build parse,
                    # even though it failed, populated the macro context
                    source = Path(Macros.expand(ref))
                    # ignore files outside of sourcedir
                    if source.parent.samefile(self.sourcedir):
                        sources.add(source.name)
            return sources

        def collect_included_sources(content):
            # collect sources included using %include
            sources = set()
            include_regex = re.compile(r"^\s*%include\s+(.*)$")
            lines = content.splitlines()
            while lines:
                line = lines.pop(0)
                m = include_regex.match(line)
                if not m:
                    continue
                arg = m.group(1)
                while line.endswith("\\"):
                    line = lines.pop(0)
                    arg = arg[:-1] + line
                # we can expand macros here because the first non-build parse,
                # even though it failed, populated the macro context
                source = Path(Macros.expand(arg))
                # ignore files outside of sourcedir
                if source.parent.samefile(self.sourcedir):
                    sources.add(source.name)
            return sources

        def collect_loaded_sources(content):
            # collect sources loaded using %{load:...}
            sources = set()
            for node in ValueParser.flatten(ValueParser.parse(content)):
                if isinstance(node, BuiltinMacro) and node.name == "load":
                    # we can expand macros here because the first non-build parse,
                    # even though it failed, populated the macro context
                    source = Path(Macros.expand(node.body))
                    # ignore files outside of sourcedir
                    if source.parent.samefile(self.sourcedir):
                        sources.add(source.name)
            return sources

        tainted = False
        try:
            # do a non-build parse first, to get a list of sources
            spec = get_rpm_spec(content, rpm.RPMSPEC_ANYARCH | rpm.RPMSPEC_FORCE)
            sources = {s for s, _, _ in spec.sources}
            non_empty_sources = set()
        except RPMException:
            if not self.force_parse:
                raise
            else:
                sources = collect_included_sources(content) | collect_loaded_sources(
                    content
                )
                non_empty_sources = collect_sources_referenced_from_tags(content)
                if not sources and not non_empty_sources:
                    # no point in trying again
                    raise
                with self._make_dummy_sources(
                    sources, non_empty_sources
                ) as dummy_sources:
                    if not dummy_sources:
                        raise
                    filelist = "\n".join(str(ds) for ds in dummy_sources)
                    logger.warning(
                        f"Created dummy sources for nonexistent files:\n{filelist}"
                    )
                    tainted = True
                    # do a non-build parse again with dummy sources
                    spec = get_rpm_spec(
                        content, rpm.RPMSPEC_ANYARCH | rpm.RPMSPEC_FORCE
                    )
                    # spec.sources contains also previously collected
                    # non empty sources (if any), remove them
                    sources = {s for s, _, _ in spec.sources} - non_empty_sources

        # workaround RPM lua tables feature/bug
        #
        # RPM lua tables are a global storage that is used by lua macros and in turn
        # also by standard %sources and %patches macros
        #
        # they are initialized when a rpm.spec instance is created and destroyed
        # when a rpm.spec instance is deleted
        #
        # when a variable that holds a rpm.spec instance is assigned a new rpm.spec
        # instance, the old instance is garbage collected afterwards and that destroys
        # lua tables that were (re)initialized when the new instance was created
        #
        # explicitly deleting the old instance before creating a new one prevents this
        del spec

        with self._make_dummy_sources(sources, non_empty_sources):
            # do a full parse with dummy sources
            return get_rpm_spec(content, rpm.RPMSPEC_ANYARCH), tainted

    def parse(
        self, content: str, extra_macros: Optional[List[Tuple[str, str]]] = None
    ) -> None:
        """
        Parses the content of a spec file and updates the `spec` and `tainted` attributes.

        Args:
            content: String representing the content of a spec file.
            extra_macros: List of extra macro definitions.

        Raises:
            RPMException, if parsing error occurs.
        """
        if self.spec:
            # workaround RPM lua tables feature/bug, see above for details
            del self.spec
        try:
            self.spec, self.tainted = self._do_parse(content, extra_macros)
        except RPMException:
            self.spec = None
            self.tainted = False
            raise



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import collections
import copy
import re
from typing import TYPE_CHECKING, List, Optional, SupportsIndex, Union, cast, overload

from specfile.constants import (
    SCRIPT_SECTIONS,
    SECTION_NAMES,
    SECTION_OPTIONS,
    SIMPLE_SCRIPT_SECTIONS,
)
from specfile.formatter import formatted
from specfile.macro_definitions import MacroDefinitions
from specfile.macros import Macros
from specfile.options import Options

if TYPE_CHECKING:
    from specfile.specfile import Specfile

# name for the implicit "preamble" section
PREAMBLE = "package"


class Section(collections.UserList):
    """
    Class that represents a spec file section.

    Attributes:
        name: Name of the section (without the leading '%').
        options: Options of the section.
        data: List of lines forming the content of the section,
          not including newline characters.
    """

    def __init__(
        self,
        name: str,
        options: Optional[Options] = None,
        delimiter: str = "",
        separator: str = "\n",
        data: Optional[List[str]] = None,
    ) -> None:
        """
        Constructs a `Section` object.

        Args:
            name: Name of the section (without the leading '%').
            options: Options of the section.
            delimiter: Delimiter separating name and option string.
            separator: String separating name and options from section content,
              defaults to newline.
            data: List of lines forming the content of the section,
              not including newline characters.

        Returns:
            Constructed instance of `Section` class.
        """
        super().__init__()
        if name.lower() not in SECTION_NAMES:
            raise ValueError(f"Invalid section name: '{name}'")
        self.name = name
        self.options = (
            copy.deepcopy(options)
            if options is not None
            else Options([], SECTION_OPTIONS.get(name.lower()))
        )
        self._delimiter = delimiter
        self._separator = separator
        if data is not None:
            self.data = data.copy()

    def __str__(self) -> str:
        data = "".join(f"{i}\n" for i in self.data)
        if self.normalized_id == PREAMBLE:
            return data
        return f"%{self.id}{self._separator}{data}"

    @formatted
    def __repr__(self) -> str:
        return (
            f"Section({self.name!r}, {self.options!r}, {self._delimiter!r}, "
            f"{self._separator!r}, {self.data!r})"
        )

    @overload
    def __getitem__(self, i: SupportsIndex) -> str:
        pass

    @overload
    def __getitem__(self, i: slice) -> "Section":
        pass

    def __getitem__(self, i):
        if isinstance(i, slice):
            return Section(
                self.name, self.options, self._delimiter, self._separator, self.data[i]
            )
        else:
            return self.data[i]

    @property
    def normalized_name(self) -> str:
        """Normalized name of the section. All characters are lowercased."""
        return self.name.lower()

    @property
    def id(self) -> str:
        """ID of the section (name and options, without the leading '%')."""
        # ensure delimiter is not empty when there are any options
        if self.options and not self._delimiter:
            self._delimiter = " "
        return self.name + self._delimiter + str(self.options)

    @property
    def normalized_id(self) -> str:
        """Normalized ID of the section. All characters of name are lowercased."""
        # ensure delimiter is not empty when there are any options
        if self.options and not self._delimiter:
            self._delimiter = " "
        return self.normalized_name + self._delimiter + str(self.options)

    @property
    def is_script(self) -> bool:
        """Whether the content of the section is a shell script."""
        normalized_name = self.normalized_id.split()[0]
        return normalized_name in SCRIPT_SECTIONS | SIMPLE_SCRIPT_SECTIONS

    def copy(self) -> "Section":
        return copy.copy(self)

    def get_raw_data(self) -> List[str]:
        if self.normalized_id == PREAMBLE:
            return self.data
        return str(self).splitlines()


class Sections(collections.UserList):
    """
    Class that represents all spec file sections, hence the entire spec file.

    Sections can be accessed by index or conveniently by id as attributes:
    ```
    # print the third line of the first section
    print(sections[0][2])

    # remove the last line of %prep section
    del sections.prep[-1]

    # replace the entire %prep section
    sections.prep = ['line 1', 'line 2']

    # delete %changelog
    del sections.changelog
    ```

    Attributes:
        data: List of individual sections. Preamble is expected to always be the first.
    """

    def __str__(self) -> str:
        return "".join(str(i) for i in self.data)

    @formatted
    def __repr__(self) -> str:
        return f"Sections({self.data!r})"

    def __contains__(self, id: object) -> bool:
        try:
            # use parent's __getattribute__() so this method can be called from __getattr__()
            data = super().__getattribute__("data")
        except AttributeError:
            return False
        return any(s.normalized_id == cast(str, id).lower() for s in data)

    def __getattr__(self, id: str) -> Section:
        if id not in self:
            return super().__getattribute__(id)
        try:
            return self.get(id)
        except ValueError:
            raise AttributeError(id)

    def __setattr__(self, id: str, value: Union[Section, List[str]]) -> None:
        if id not in self:
            return super().__setattr__(id, value)
        try:
            if isinstance(value, Section):
                self.data[self.find(id)] = value
            else:
                self.data[self.find(id)].data = value
        except ValueError:
            raise AttributeError(id)

    def __delattr__(self, id: str) -> None:
        if id not in self:
            return super().__delattr__(id)
        try:
            del self.data[self.find(id)]
        except ValueError:
            raise AttributeError(id)

    def copy(self) -> "Sections":
        return copy.copy(self)

    def get(self, id: str) -> Section:
        return self.data[self.find(id)]

    def find(self, id: str) -> int:
        for i, section in enumerate(self.data):
            if section.normalized_id == id.lower():
                return i
        raise ValueError

    @classmethod
    def parse(
        cls, lines: List[str], context: Optional["Specfile"] = None
    ) -> "Sections":
        """
        Parses given lines into sections.

        Args:
            lines: Lines to parse.
            context: `Specfile` instance that defines the context for macro expansions.

        Returns:
            Constructed instance of `Sections` class.
        """

        def expand(s):
            if context:
                result = context.expand(
                    s, skip_parsing=getattr(expand, "skip_parsing", False)
                )
                # parse only once
                expand.skip_parsing = True
                return result
            return Macros.expand(s)

        def split_id(line):
            content = []
            separator = "\n"
            tokens = re.split(r"(\s+)", line)
            if len(tokens) > 2:
                # if the last token after macro expansion starts with a newline,
                # consider it part of section content
                if expand(tokens[-1]).startswith("\n"):
                    content = [tokens.pop()]
                    separator = tokens.pop()
            if len(tokens) > 2:
                name = tokens[0]
                delimiter = tokens[1]
                options = Options(
                    Options.tokenize("".join(tokens[2:])),
                    SECTION_OPTIONS.get(name.lower()),
                )
                return name, options, delimiter, separator, content
            return tokens[0], None, "", separator, content

        excluded_lines = []
        macro_definitions = MacroDefinitions.parse(lines)
        for md in macro_definitions:
            position = md.get_position(macro_definitions)
            excluded_lines.append(range(position, position + len(md.body.splitlines())))
        section_id_regexes = [
            re.compile(rf"^%{re.escape(n)}(\s+.*(?<!\\)$|$)", re.IGNORECASE)
            for n in SECTION_NAMES
        ]
        section_starts = []
        for i, line in enumerate(lines):
            # section can not start inside macro definition body
            if any(i in r for r in excluded_lines):
                continue
            if line.startswith("%"):
                for r in section_id_regexes:
                    if r.match(line):
                        section_starts.append(i)
                        break
        section_starts.append(len(lines))
        data = [Section(PREAMBLE, data=lines[: section_starts[0]])]
        for start, end in zip(section_starts, section_starts[1:]):
            name, options, delimiter, separator, content = split_id(lines[start][1:])
            data.append(
                Section(
                    name,
                    options,
                    delimiter,
                    separator,
                    content + lines[start + 1 : end],
                )
            )
        return cls(data)

    def get_raw_data(self) -> List[str]:
        result = []
        for section in self.data:
            result.extend(section.get_raw_data())
        return result



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import collections
import copy
import datetime
import re
from typing import List, Optional, SupportsIndex, Union, overload

import rpm

from specfile.exceptions import SpecfileException
from specfile.formatter import formatted
from specfile.sections import Section
from specfile.utils import EVR

WEEKDAYS = ("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun")
MONTHS = (
    "Jan",
    "Feb",
    "Mar",
    "Apr",
    "May",
    "Jun",
    "Jul",
    "Aug",
    "Sep",
    "Oct",
    "Nov",
    "Dec",
)


class ChangelogEntry:
    """
    Class that represents a changelog entry.

    Attributes:
        header: Header of the entry.
        content: List of lines forming the content of the entry.
    """

    def __init__(
        self,
        header: str,
        content: List[str],
        following_lines: Optional[List[str]] = None,
    ) -> None:
        """
        Constructs a `ChangelogEntry` object.

        Args:
            header: Header of the entry.
            content: List of lines forming the content of the entry.
            following_lines: Extra lines that follow the entry.

        Returns:
            Constructed instance of `ChangelogEntry` class.
        """
        self.header = header
        self.content = content.copy()
        self._following_lines = (
            following_lines.copy() if following_lines is not None else []
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ChangelogEntry):
            return NotImplemented
        return (
            self.header == other.header
            and self.content == other.content
            and self._following_lines == other._following_lines
        )

    def __str__(self) -> str:
        return f"{self.header}\n" + "\n".join(self.content) + "\n"

    @formatted
    def __repr__(self) -> str:
        return f"ChangelogEntry({self.header!r}, {self.content!r}, {self._following_lines!r})"

    @property
    def evr(self) -> Optional[str]:
        """EVR (Epoch, Version, Release) of the entry."""
        m = re.match(
            r"""
            ^.*
            \s+                       # preceding whitespace
            ((?P<sb>\[)|(?P<rb>\())?  # optional opening bracket
            (?P<evr>(\d+:)?\S+-\S+?)  # EVR
            (?(sb)\]|(?(rb)\)))       # matching closing bracket
            :?                        # optional colon
            \s*                       # optional following whitespace
            $
            """,
            self.header,
            re.VERBOSE,
        )
        if not m:
            return None
        return m.group("evr")

    @property
    def extended_timestamp(self) -> bool:
        """Whether the timestamp present in the entry header is extended (date and time)."""
        weekdays = "|".join(WEEKDAYS)
        months = "|".join(MONTHS)
        m = re.search(
            rf"""
            ({weekdays})     # weekday
            [ ]
            ({months})       # month
            [ ]+
            ([12]?\d|3[01])  # day of month
            [ ]
            ([01]\d|2[0-3])  # hour
            :
            ([0-5]\d)        # minute
            :
            ([0-5]\d)        # second
            [ ]
            \S+              # timezone
            [ ]
            \d{{4}}          # year
            """,
            self.header,
            re.VERBOSE,
        )
        return m is not None

    @property
    def day_of_month_padding(self) -> str:
        """Padding of day of month in the entry header timestamp"""
        weekdays = "|".join(WEEKDAYS)
        months = "|".join(MONTHS)
        m = re.search(
            rf"""
            ({weekdays})                 # weekday
            [ ]
            ({months})                   # month
            [ ]
            (?P<wsp>[ ]*)                # optional whitespace padding
            ((?P<zp>0)?\d|[12]\d|3[01])  # possibly zero-padded day of month
            """,
            self.header,
            re.VERBOSE,
        )
        if not m:
            return ""
        return m.group("wsp") + (m.group("zp") or "")

    @classmethod
    def assemble(
        cls,
        timestamp: Union[datetime.date, datetime.datetime],
        author: str,
        content: List[str],
        evr: Optional[str] = None,
        day_of_month_padding: str = "0",
        append_newline: bool = True,
    ) -> "ChangelogEntry":
        """
        Assembles a changelog entry.

        Args:
            timestamp: Timestamp of the entry.
              Supply `datetime` rather than `date` for extended format.
            author: Author of the entry.
            content: List of lines forming the content of the entry.
            evr: EVR (epoch, version, release) of the entry.
            day_of_month_padding: Padding to apply to day of month in the timestamp.
            append_newline: Whether the entry should be followed by an empty line.

        Returns:
            Constructed instance of `ChangelogEntry` class.
        """
        weekday = WEEKDAYS[timestamp.weekday()]
        month = MONTHS[timestamp.month - 1]
        header = f"* {weekday} {month}"
        if day_of_month_padding.endswith("0"):
            header += f" {day_of_month_padding[:-1]}{timestamp.day:02}"
        else:
            header += f" {day_of_month_padding}{timestamp.day}"
        if isinstance(timestamp, datetime.datetime):
            # extended format
            if not timestamp.tzinfo:
                timestamp = timestamp.replace(tzinfo=datetime.timezone.utc)
            header += f" {timestamp:%H:%M:%S %Z}"
        header += f" {timestamp:%Y} {author}"
        if evr is not None:
            header += f" - {evr}"
        return cls(header, content, [""] if append_newline else None)


class Changelog(collections.UserList):
    """
    Class that represents a changelog.

    Attributes:
        data: List of individual entries.
    """

    def __init__(
        self,
        data: Optional[List[ChangelogEntry]] = None,
        predecessor: Optional[List[str]] = None,
    ) -> None:
        """
        Constructs a `Changelog` object.

        Args:
            data: List of individual changelog entries.
            predecessor: Lines at the beginning of a section that can't be parsed
              into changelog entries.

        Returns:
            Constructed instance of `Changelog` class.
        """
        super().__init__()
        if data is not None:
            self.data = data.copy()
        self._predecessor = predecessor.copy() if predecessor is not None else []

    def __str__(self) -> str:
        return "\n".join(str(i) for i in reversed(self.data))

    @formatted
    def __repr__(self) -> str:
        return f"Changelog({self.data!r}, {self._predecessor!r})"

    @overload
    def __getitem__(self, i: SupportsIndex) -> ChangelogEntry:
        pass

    @overload
    def __getitem__(self, i: slice) -> "Changelog":
        pass

    def __getitem__(self, i):
        if isinstance(i, slice):
            return Changelog(self.data[i], self._predecessor)
        else:
            return self.data[i]

    def __delitem__(self, i: Union[SupportsIndex, slice]) -> None:
        def delete(index):
            following_lines = self.data[index]._following_lines.copy()
            del self.data[index]
            if index < len(self.data):
                self.data[index]._following_lines.extend(following_lines[1:])
            else:
                self._predecessor.extend(following_lines[1:])

        if isinstance(i, slice):
            for index in reversed(range(len(self.data))[i]):
                delete(index)
        else:
            delete(i)

    def copy(self) -> "Changelog":
        return copy.copy(self)

    def filter(
        self, since: Optional[str] = None, until: Optional[str] = None
    ) -> "Changelog":
        """
        Filters changelog entries with EVR between since and until.

        Args:
            since: Optional lower bound. If specified, entries with EVR higher
              than or equal to this will be included.
            until: Optional upper bound. If specified, entries with EVR lower
              than or equal to this will be included.

        Returns:
            Filtered changelog.
        """

        def parse_evr(s):
            try:
                evr = EVR.from_string(s)
            except SpecfileException:
                return "0", "0", ""
            return str(evr.epoch), evr.version or "0", evr.release

        if since is None:
            start_index = 0
        else:
            start_index = next(
                (
                    i
                    for i, e in enumerate(self.data)
                    if rpm.labelCompare(parse_evr(e.evr), parse_evr(since)) >= 0
                ),
                len(self.data) + 1,
            )
        if until is None:
            end_index = len(self.data) + 1
        else:
            end_index = next(
                (
                    i + 1
                    for i, e in reversed(list(enumerate(self.data)))
                    if rpm.labelCompare(parse_evr(e.evr), parse_evr(until)) <= 0
                ),
                0,
            )
        return self[start_index:end_index]

    @classmethod
    def parse(cls, section: Section) -> "Changelog":
        """
        Parses a %changelog section.

        Args:
            section: Section to parse.

        Returns:
            Constructed instance of `Changelog` class.
        """

        def extract_following_lines(content):
            following_lines: List[str] = []
            while content and not content[-1].strip():
                following_lines.insert(0, content.pop())
            return following_lines

        data: List[ChangelogEntry] = []
        predecessor = []
        header = None
        content: List[str] = []
        for line in section:
            if line.startswith("*"):
                if header is None or "".join(content).strip():
                    if header:
                        following_lines = extract_following_lines(content)
                        data.insert(0, ChangelogEntry(header, content, following_lines))
                    header = line
                    content = []
                else:
                    content.append(line)
            elif header:
                content.append(line)
            else:
                predecessor.append(line)
        if header:
            following_lines = extract_following_lines(content)
            data.insert(0, ChangelogEntry(header, content, following_lines))
        return cls(data, predecessor)

    def get_raw_section_data(self) -> List[str]:
        """
        Reconstructs section data from changelog.

        Returns:
            List of lines forming the reconstructed section data.
        """
        result = self._predecessor.copy()
        for entry in reversed(self.data):
            result.append(entry.header)
            result.extend(entry.content)
            result.extend(entry._following_lines)
        return result



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import collections
import re
import urllib.parse
from abc import ABC, abstractmethod
from typing import Iterable, List, Optional, Tuple, Union, cast, overload

from specfile.exceptions import DuplicateSourceException
from specfile.formatter import formatted
from specfile.sourcelist import Sourcelist, SourcelistEntry
from specfile.tags import Comments, Tag, Tags
from specfile.utils import get_filename_from_location


class Source(ABC):
    """Class that represents a source."""

    @property
    @abstractmethod
    def number(self) -> int:
        """Source number."""
        ...

    @property
    @abstractmethod
    def location(self) -> str:
        """Literal location of the source as stored in the spec file."""
        ...

    @location.setter
    def location(self, value: str) -> None:
        ...

    @property
    @abstractmethod
    def expanded_location(self) -> Optional[str]:
        """Location of the source after expanding macros."""
        ...

    @property
    @abstractmethod
    def filename(self) -> str:
        """Literal filename of the source."""
        ...

    @property
    def remote(self) -> bool:
        """Whether the source is remote (location is URL)."""
        url = urllib.parse.urlsplit(self.expanded_location)
        return all((url.scheme, url.netloc))

    @property
    @abstractmethod
    def expanded_filename(self) -> Optional[str]:
        """Filename of the source after expanding macros."""
        ...

    @property
    @abstractmethod
    def comments(self) -> Comments:
        """List of comments associated with the source."""
        ...


class TagSource(Source):
    """Class that represents a source backed by a spec file tag."""

    def __init__(self, tag: Tag, number: Optional[int] = None) -> None:
        """
        Constructs a `TagSource` object.

        Args:
            tag: Tag that this source represents.
            number: Source number (in the case of implicit numbering).

        Returns:
            Constructed instance of `TagSource` class.
        """
        self._tag = tag
        self._number = number

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TagSource):
            return NotImplemented
        return self._tag == other._tag and self._number == other._number

    @formatted
    def __repr__(self) -> str:
        # determine class name dynamically so that inherited classes
        # don't have to reimplement __repr__()
        return f"{self.__class__.__name__}({self._tag!r}, {self._number!r})"

    def _extract_number(self) -> Optional[str]:
        """
        Extracts source number from tag name.

        Returns:
            Extracted number or None if there isn't one.
        """
        tokens = re.split(r"(\d+)", self._tag.name, maxsplit=1)
        if len(tokens) > 1:
            return tokens[1]
        return None

    @property
    def number(self) -> int:
        """Source number."""
        if self._number is not None:
            return self._number
        return int(self._extract_number() or 0)

    @property
    def number_digits(self) -> int:
        """
        Gets number of digits in the source number.

        Returns 0 if the source has no number, 1 if the source number
        has no leading zeros and the actual number of digits if there are
        any leading zeros.
        """
        if self._number is not None:
            return 0
        number = self._extract_number()
        if not number:
            return 0
        return len(number) if number.startswith("0") else 1

    @property
    def location(self) -> str:
        """Literal location of the source as stored in the spec file."""
        return self._tag.value

    @location.setter
    def location(self, value: str) -> None:
        self._tag.value = value

    @property
    def expanded_location(self) -> Optional[str]:
        """Location of the source after expanding macros."""
        return self._tag.expanded_value

    @property
    def filename(self) -> str:
        """Literal filename of the source."""
        return get_filename_from_location(self._tag.value)

    @property
    def expanded_filename(self) -> Optional[str]:
        """Filename of the source after expanding macros."""
        if self._tag.expanded_value is None:
            return None
        return get_filename_from_location(self._tag.expanded_value)

    @property
    def comments(self) -> Comments:
        """List of comments associated with the source."""
        return self._tag.comments


class ListSource(Source):
    """Class that represents a source backed by a line in a %sourcelist section."""

    def __init__(self, source: SourcelistEntry, number: int) -> None:
        """
        Constructs a `ListSource` object.

        Args:
            source: Sourcelist entry that this source represents.
            number: Source number.

        Returns:
            Constructed instance of `ListSource` class.
        """
        self._source = source
        self._number = number

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ListSource):
            return NotImplemented
        return self._source == other._source and self._number == other._number

    @formatted
    def __repr__(self) -> str:
        # determine class name dynamically so that inherited classes
        # don't have to reimplement __repr__()
        return f"{self.__class__.__name__}({self._source!r}, {self._number!r})"

    @property
    def number(self) -> int:
        """Source number."""
        return self._number

    @property
    def location(self) -> str:
        """Literal location of the source as stored in the spec file."""
        return self._source.location

    @location.setter
    def location(self, value: str) -> None:
        self._source.location = value

    @property
    def expanded_location(self) -> str:
        """Location of the source after expanding macros."""
        return self._source.expanded_location

    @property
    def filename(self) -> str:
        """Literal filename of the source."""
        return get_filename_from_location(self._source.location)

    @property
    def expanded_filename(self) -> str:
        """Filename of the source after expanding macros."""
        return get_filename_from_location(self._source.expanded_location)

    @property
    def comments(self) -> Comments:
        """List of comments associated with the source."""
        return self._source.comments


class Sources(collections.abc.MutableSequence):
    """Class that represents a sequence of all sources."""

    prefix: str = "Source"
    tag_class: type = TagSource
    list_class: type = ListSource

    def __init__(
        self,
        tags: Tags,
        sourcelists: List[Sourcelist],
        allow_duplicates: bool = False,
        default_to_implicit_numbering: bool = False,
        default_source_number_digits: int = 1,
    ) -> None:
        """
        Constructs a `Sources` object.

        Args:
            tags: All spec file tags.
            sourcelists: List of all %sourcelist sections.
            allow_duplicates: Whether to allow duplicate entries when adding new sources.
            default_to_implicit_numbering: Use implicit numbering (no source numbers) by default.
            default_source_number_digits: Default number of digits in a source number.
            context: `Specfile` instance that defines the context for macro expansions.

        Returns:
            Constructed instance of `Sources` class.
        """
        self._tags = tags
        self._sourcelists = sourcelists
        self._allow_duplicates = allow_duplicates
        self._default_to_implicit_numbering = default_to_implicit_numbering
        self._default_source_number_digits = default_source_number_digits

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Sources):
            return NotImplemented
        return (
            self._tags == other._tags
            and self._sourcelists == other._sourcelists
            and self._allow_duplicates == other._allow_duplicates
            and self._default_to_implicit_numbering
            == other._default_to_implicit_numbering
            and self._default_source_number_digits
            == other._default_source_number_digits
        )

    @formatted
    def __repr__(self) -> str:
        # determine class name dynamically so that inherited classes
        # don't have to reimplement __repr__()
        return (
            f"{self.__class__.__name__}({self._tags!r}, {self._sourcelists!r}, "
            f"{self._allow_duplicates!r}, {self._default_to_implicit_numbering!r}, "
            f"{self._default_source_number_digits!r})"
        )

    def __contains__(self, location: object) -> bool:
        items = self._get_items()
        if not items:
            return False
        return location in [s.location for s in list(zip(*items))[0]]

    def __add__(
        self, other: Union[Source, Iterable[Source], "Sources"]
    ) -> List[Source]:
        if isinstance(other, Source):
            return list(self) + [other]
        return list(self) + list(other)

    def __len__(self) -> int:
        return len(self._get_items())

    @overload
    def __getitem__(self, i: int) -> Source:
        pass

    @overload
    def __getitem__(self, i: slice) -> List[Source]:
        pass

    def __getitem__(self, i):
        items = self._get_items()
        if isinstance(i, slice):
            return list(zip(*items[i]))[0]
        else:
            return items[i][0]

    @overload
    def __setitem__(self, i: int, item: str) -> None:
        pass

    @overload
    def __setitem__(self, i: slice, item: Iterable[str]) -> None:
        pass

    def __setitem__(self, i, item):
        items = self._get_items()
        if isinstance(i, slice):
            for i0, i1 in enumerate(range(len(items))[i]):
                items[i1][0].location = item[i0]
        else:
            items[i][0].location = item

    def __delitem__(self, i: Union[int, slice]) -> None:
        items = self._get_items()
        if isinstance(i, slice):
            for _, container, index in reversed(items[i]):
                del container[index]
        else:
            _, container, index = items[i]
            del container[index]

    def _get_tags(self) -> List[Tuple[TagSource, Tags, int]]:
        """
        Gets all tag sources.

        Returns:
            List of tuples in the form of (source, container, index),
            where source is an instance of `TagSource` representing a tag,
            container is the container the tag is part of and index
            is its index within that container.
        """
        result = []
        last_number = -1
        for i, tag in enumerate(self._tags):
            if tag.normalized_name == self.prefix:
                last_number += 1
                ts = self.tag_class(tag, last_number)
            elif tag.normalized_name.startswith(self.prefix):
                ts = self.tag_class(tag)
                last_number = ts.number
            else:
                continue
            result.append((ts, self._tags, i))
        return result

    def _get_items(self) -> List[Tuple[Source, Union[Tags, Sourcelist], int]]:
        """
        Gets all sources.

        Returns:
            List of tuples in the form of (source, container, index),
            where source is an instance of `TagSource` or `ListSource`
            representing a source, container is the container the source
            is part of and index is its index within that container.
        """
        result = cast(
            List[Tuple[Source, Union[Tags, Sourcelist], int]], self._get_tags()
        )
        last_number = result[-1][0].number if result else -1
        result.extend(
            (self.list_class(sl[i], last_number + 1 + i), sl, i)
            for sl in self._sourcelists
            for i in range(len(sl))
        )
        return result

    def _detect_implicit_numbering(self) -> bool:
        """
        Tries to detect if implicit numbering is being used, i.e. Source/Patch
        tags don't have numbers.

        Returns:
            True if implicit numbering is being/should be used, False otherwise.
        """
        tags = self._get_tags()
        if any(t._number is None for t, _, _ in tags):
            return False
        if len(tags) <= 1:
            return self._default_to_implicit_numbering
        return True

    def _get_tag_format(
        self,
        reference: TagSource,
        number: int,
        number_digits_override: Optional[int] = None,
    ) -> Tuple[str, str]:
        """
        Determines name and separator of a new source tag based on
        a reference tag and the requested source number.

        The new name has the same number of digits as the reference
        (unless number_digits_override is set to a different value)
        and the length of the separator is adjusted accordingly.

        Args:
            reference: Reference tag source.
            number: Requested source number.
            number_digits_override: Requested number of digits in the source number.

        Returns:
            Tuple in the form of (name, separator).
        """
        if number_digits_override is not None:
            number_digits = number_digits_override
        else:
            number_digits = reference.number_digits
        if self._detect_implicit_numbering() or number_digits == 0:
            suffix = ""
        else:
            suffix = f"{number:0{number_digits}}"
        name = f"{self.prefix}{suffix}"
        diff = len(reference._tag.name) - len(name)
        if diff >= 0:
            return name, reference._tag._separator + " " * diff
        return name, reference._tag._separator[:diff] or ":"

    def _get_initial_tag_setup(self, number: int = 0) -> Tuple[int, str, str]:
        """
        Determines the initial placement, name and separator of
        a new source tag. The placement is expressed as an index
        in the list of all tags.

        Args:
            number: Initial source number, defaults to 0.

        Returns:
            Tuple in the form of (index, name, separator).
        """
        if (
            self._default_to_implicit_numbering
            or self._default_source_number_digits == 0
        ):
            suffix = ""
        else:
            suffix = f"{number:0{self._default_source_number_digits}}"
        return len(self._tags) if self._tags else 0, f"{self.prefix}{suffix}", ": "

    def _deduplicate_tag_names(self, start: int = 0) -> None:
        """
        Eliminates duplicate numbers in source tag names.

        Args:
            start: Starting index, defaults to the first source tag.
        """
        tags = self._get_tags()
        if not tags:
            return
        tag_sources = list(zip(*tags[start:]))[0]
        for ts0, ts1 in zip(tag_sources, tag_sources[1:]):
            if ts1.number == ts0.number:
                if ts1._number is not None:
                    ts1._number = ts0.number + 1
                else:
                    ts1._tag.name, ts1._tag._separator = self._get_tag_format(
                        ts1, ts0.number + 1
                    )

    def insert(self, i: int, location: str) -> None:
        """
        Inserts a new source at a specified index.

        Args:
            i: Requested index.
            location: Location of the new source.

        Raises:
            DuplicateSourceException if duplicates are disallowed and there
              already is a source with the same location.
        """
        if not self._allow_duplicates and location in self:
            raise DuplicateSourceException(f"{self.prefix} '{location}' already exists")
        items = self._get_items()
        if i > len(items):
            i = len(items)
        if items:
            if i == len(items):
                source, container, index = items[-1]
                index += 1
                number = source.number + 1
            else:
                source, container, index = items[i]
                number = source.number
            if isinstance(source, self.tag_class):
                name, separator = self._get_tag_format(cast(TagSource, source), number)
                container.insert(
                    index,
                    Tag(name, location, separator, Comments()),
                )
                self._deduplicate_tag_names(i)
            else:
                container.insert(
                    index,
                    SourcelistEntry(location, Comments()),  # type: ignore[arg-type]
                )
        elif self._sourcelists:
            self._sourcelists[-1].append(SourcelistEntry(location, Comments()))
        else:
            index, name, separator = self._get_initial_tag_setup()
            self._tags.insert(
                index,
                Tag(name, location, separator, Comments()),
            )

    def insert_numbered(self, number: int, location: str) -> int:
        """
        Inserts a new source with the specified number.

        Args:
            number: Number of the new source.
            location: Location of the new source.

        Returns:
            Index of the newly inserted source.

        Raises:
            DuplicateSourceException if duplicates are disallowed and there
              already is a source with the same location.
        """
        if not self._allow_duplicates and location in self:
            raise DuplicateSourceException(f"{self.prefix} '{location}' already exists")
        tags = self._get_tags()
        if tags:
            # find the nearest source tag
            i, (source, _, index) = min(
                enumerate(tags), key=lambda t: abs(t[1][0].number - number)
            )
            if source.number < number:
                i += 1
                index += 1
            name, separator = self._get_tag_format(source, number)
        else:
            i = 0
            index, name, separator = self._get_initial_tag_setup(number)
        self._tags.insert(index, Tag(name, location, separator, Comments()))
        self._deduplicate_tag_names(i)
        return i

    def remove(self, location: str) -> None:
        """
        Removes sources by location.

        Args:
            location: Location of the sources to be removed.
        """
        for source, container, index in reversed(self._get_items()):
            if source.location == location:
                del container[index]

    def remove_numbered(self, number: int) -> None:
        """
        Removes a source by number.

        Args:
            number: Number of the source to be removed.
        """
        items = self._get_items()
        try:
            container, index = next((c, i) for s, c, i in items if s.number == number)
        except StopIteration:
            pass
        else:
            del container[index]

    def count(self, location: str) -> int:
        """
        Counts sources by location.

        Args:
            location: Location of the sources to be counted.

        Returns:
            Number of sources with the specified location.
        """
        items = self._get_items()
        if not items:
            return 0
        return len([s for s in list(zip(*items))[0] if s.location == location])


class Patch(Source):
    """Class that represents a patch."""


class TagPatch(TagSource, Patch):
    """Class that represents a patch backed by a spec file tag."""


class ListPatch(ListSource, Patch):
    """Class that represents a patch backed by a line in a %patchlist section."""


class Patches(Sources):
    """Class that represents a sequence of all patches."""

    prefix: str = "Patch"
    tag_class: type = TagPatch
    list_class: type = ListPatch

    def _get_initial_tag_setup(self, number: int = 0) -> Tuple[int, str, str]:
        """
        Determines the initial placement, name and separator of
        a new source tag. The placement is expressed as an index
        in the list of all tags.

        Args:
            number: Initial source number, defaults to 0.

        Returns:
            Tuple in the form of (index, name, separator).
        """
        try:
            index, source = [
                (i, Sources.tag_class(t))
                for i, t in enumerate(self._tags)
                if t.normalized_name.startswith(Sources.prefix)
            ][-1]
        except IndexError:
            return super()._get_initial_tag_setup(number)
        name, separator = self._get_tag_format(
            source, number, self._default_source_number_digits
        )
        return index + 1, name, separator



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import copy
import datetime
import subprocess

import pytest
import rpm
from flexmock import flexmock

from specfile.exceptions import RPMException, SpecfileException
from specfile.prep import AutopatchMacro, AutosetupMacro, PatchMacro, SetupMacro
from specfile.sections import Section
from specfile.specfile import Specfile


def test_parse(spec_multiple_sources):
    spec = Specfile(spec_multiple_sources)
    prep = spec.rpm_spec.prep
    # remove all sources
    for path in spec.sourcedir.iterdir():
        if not path.samefile(spec.path):
            path.unlink()
    spec = Specfile(spec_multiple_sources)
    assert spec.rpm_spec.prep == prep


def test_prep_traditional(spec_traditional):
    spec = Specfile(spec_traditional)
    with spec.prep() as prep:
        assert AutosetupMacro not in prep.macros
        assert AutopatchMacro not in prep.macros
        assert isinstance(prep.macros[0], SetupMacro)
        assert prep.macros[0] == prep.setup
        for i, m in enumerate(prep.macros[1:]):
            assert isinstance(m, PatchMacro)
            assert m.number == i
            assert m.options.p == 1
        prep.remove_patch_macro(0)
        assert len([m for m in prep.macros if isinstance(m, PatchMacro)]) == 2
        prep.add_patch_macro(0, p=2, b=".test")
        assert len(prep.macros) == 4
        assert prep.patch0.options.p == 2
        assert prep.patch0.options.b == ".test"
        prep.patch0.options.b = ".test2"
        prep.patch0.options.E = True
    with spec.sections() as sections:
        assert sections.prep[1] == "%patch0 -p2 -b .test2 -E"


def test_prep_autosetup(spec_autosetup):
    spec = Specfile(spec_autosetup)
    with spec.prep() as prep:
        assert len(prep.macros) == 1
        assert AutosetupMacro in prep.macros
        assert SetupMacro not in prep.macros
        assert PatchMacro not in prep.macros
        assert prep.autosetup.options.p == 1


def test_prep_autopatch(spec_autopatch):
    spec = Specfile(spec_autopatch)
    with spec.prep() as prep:
        assert len(prep.macros) == 4
        assert prep.macros[1].options.M == 2
        assert prep.macros[2].options.positional == [3]
        assert prep.macros[3].options.m == 4
        del prep.macros[1]
        del prep.macros[2]
        prep.autopatch.options.positional = list(range(7))
    with spec.sections() as sections:
        assert sections.prep[0] == "%autosetup -N"
        assert sections.prep[3] == "%autopatch -p1 0 1 2 3 4 5 6"


def test_sources(spec_minimal):
    spec = Specfile(spec_minimal)
    source = "test.tar.gz"
    with spec.sources() as sources:
        assert not sources
        sources.append(source)
        assert sources.count(source) == len(sources) == 1
    with spec.tags() as tags:
        assert [source] == [t.value for t in tags if t.name.startswith("Source")]
    with spec.sources() as sources:
        sources.remove(source)
        assert not sources
        sources.insert(0, source)
        assert sources[0].location == source
        sources.clear()
        assert not sources


def test_patches(spec_patchlist):
    spec = Specfile(spec_patchlist)
    patch = "test.patch"
    with spec.patches() as patches:
        patches.insert(0, patch)
        assert patches[0].location == patch
        assert patches[1].number == 1
    with spec.tags() as tags:
        assert len([t for t in tags if t.name.startswith("Patch")]) == 2
    with spec.patches() as patches:
        patches.remove(patch)
        patches.insert(1, patch)
        patches[1].comments.append("test")
    with spec.sections() as sections:
        assert len([sl for sl in sections.patchlist if sl]) == 4
        assert sections.patchlist[0] == "# test"


@pytest.mark.parametrize(
    "rpmdev_packager_available, entry, author, email, timestamp, evr, result",
    [
        (False, None, None, None, None, None, None),
        (
            True,
            "test",
            None,
            None,
            datetime.date(2022, 2, 1),
            None,
            Section(
                "changelog",
                data=["* Tue Feb 01 2022 John Doe <john@doe.net> - 0.1-1", "test"],
            ),
        ),
        (
            True,
            "test",
            None,
            None,
            datetime.date(2022, 2, 1),
            "%{version}-%{release}",
            Section(
                "changelog",
                data=["* Tue Feb 01 2022 John Doe <john@doe.net> - 0.1-1", "test"],
            ),
        ),
        (
            True,
            "test",
            None,
            None,
            datetime.date(2022, 2, 1),
            "0.1-1",
            Section(
                "changelog",
                data=["* Tue Feb 01 2022 John Doe <john@doe.net> - 0.1-1", "test"],
            ),
        ),
        (
            True,
            "test",
            None,
            None,
            datetime.date(2022, 2, 1),
            "0.2-1.1",
            Section(
                "changelog",
                data=["* Tue Feb 01 2022 John Doe <john@doe.net> - 0.2-1.1", "test"],
            ),
        ),
        (
            True,
            "test",
            "Bill Packager",
            None,
            datetime.date(2022, 2, 1),
            None,
            Section(
                "changelog", data=["* Tue Feb 01 2022 Bill Packager - 0.1-1", "test"]
            ),
        ),
        (
            True,
            "test",
            "Bill Packager",
            "bill@packager.net",
            datetime.date(2022, 2, 1),
            None,
            Section(
                "changelog",
                data=[
                    "* Tue Feb 01 2022 Bill Packager <bill@packager.net> - 0.1-1",
                    "test",
                ],
            ),
        ),
        (
            True,
            "test",
            "Bill Packager",
            "bill@packager.net",
            datetime.datetime(2022, 2, 1, 9, 28, 13),
            None,
            Section(
                "changelog",
                data=[
                    "* Tue Feb 01 09:28:13 UTC 2022 Bill Packager <bill@packager.net> - 0.1-1",
                    "test",
                ],
            ),
        ),
        (
            True,
            ["line 1", "line 2"],
            "Bill Packager",
            "bill@packager.net",
            datetime.datetime(2022, 2, 1, 9, 28, 13),
            None,
            Section(
                "changelog",
                data=[
                    "* Tue Feb 01 09:28:13 UTC 2022 Bill Packager <bill@packager.net> - 0.1-1",
                    "line 1",
                    "line 2",
                ],
            ),
        ),
    ],
)
def test_add_changelog_entry(
    spec_minimal,
    rpmdev_packager_available,
    entry,
    author,
    email,
    timestamp,
    evr,
    result,
):
    if not rpmdev_packager_available:
        flexmock(subprocess).should_receive("check_output").with_args(
            "rpmdev-packager"
        ).and_raise(FileNotFoundError)
    elif author is None:
        flexmock(subprocess).should_receive("check_output").with_args(
            "rpmdev-packager"
        ).and_return(b"John Doe <john@doe.net>")
    spec = Specfile(spec_minimal)
    if not rpmdev_packager_available:
        with pytest.raises(SpecfileException):
            spec.add_changelog_entry(entry, author, email, timestamp, evr)
    else:
        spec.add_changelog_entry(entry, author, email, timestamp, evr)
        with spec.sections() as sections:
            assert sections.changelog[: len(result)] == result


@pytest.mark.parametrize(
    "version, release",
    [
        ("0.2", "3"),
        ("67", "1"),
        ("1.4.6", "0.1rc5"),
    ],
)
def test_set_version_and_release(spec_minimal, version, release):
    spec = Specfile(spec_minimal)
    spec.set_version_and_release(version, release)
    assert spec.version == version
    assert spec.release == release
    assert spec.raw_release.startswith(release)
    with spec.tags() as tags:
        assert tags.version.value == spec.version
        assert tags.release.value == spec.raw_release
    assert spec.rpm_spec.sourceHeader[rpm.RPMTAG_VERSION] == spec.expanded_version
    assert spec.rpm_spec.sourceHeader[rpm.RPMTAG_RELEASE] == spec.expanded_raw_release
    spec.raw_release = release
    with spec.tags() as tags:
        assert tags.release.value == release
    assert spec.rpm_spec.sourceHeader[rpm.RPMTAG_RELEASE] == spec.expanded_raw_release


@pytest.mark.parametrize(
    "location, number, comment",
    [
        ("patchX.patch", None, None),
        ("patchX.patch", 0, None),
        ("patch2.patch", None, None),
        ("patch3.patch", 3, "patch3"),
    ],
)
def test_add_patch(spec_autosetup, location, number, comment):
    spec = Specfile(spec_autosetup)
    if number == 0 or location == "patch2.patch":
        with pytest.raises(SpecfileException):
            spec.add_patch(location, number, comment)
    else:
        spec.add_patch(location, number, comment)
        with spec.patches() as patches:
            assert patches[-1].location == location
            if number is not None:
                assert patches[-1].number == number
            else:
                assert patches[-1].number == 3
        with spec.sections() as sections:
            if comment is not None:
                assert sections.package[-4] == f"# {comment}"


def test_remove_patches(spec_commented_patches):
    spec = Specfile(spec_commented_patches)
    with spec.patches() as patches:
        del patches[1:3]
        patches.remove_numbered(5)
    with spec.sections() as sections:
        assert sections.package[-11:-2] == [
            "# this is a downstream-only patch",
            "Patch0:         patch0.patch",
            "",
            "# these two patches are related to each other",
            "Patch3:         patch3.patch",
            "Patch4:         patch4.patch",
            "",
            "# this is patch6",
            "Patch6:         patch6.patch",
        ]


@pytest.mark.skipif(
    rpm.__version__ < "4.16", reason="%autochangelog requires rpm 4.16 or higher"
)
def test_autochangelog(spec_rpmautospec):
    spec = Specfile(spec_rpmautospec)
    assert spec.has_autochangelog
    with spec.changelog() as changelog:
        assert len(changelog) == 0
    with spec.sections() as sections:
        changelog = sections.changelog.copy()
    spec.add_changelog_entry("test")
    with spec.sections() as sections:
        assert sections.changelog == changelog


def test_update_tag(spec_macros):
    spec = Specfile(spec_macros)
    spec.update_tag("Version", "1.2.3~beta4")
    with spec.macro_definitions() as md:
        assert md.majorver.body == "1"
        assert md.minorver.body == "2"
        assert md.patchver.body == "3"
        assert md.prever.body == "beta4"
        assert md.package_version.body == "%{majorver}.%{minorver}.%{patchver}"
    assert spec.version == "%{package_version}%{?prever:~%{prever}}"
    spec.update_tag("Version", "4.0~alpha1")
    with spec.macro_definitions() as md:
        assert md.majorver.body == "1"
        assert md.minorver.body == "2"
        assert md.patchver.body == "3"
        assert md.prever.body == "alpha1"
        assert md.package_version.body == "4.0"
    assert spec.version == "%{package_version}%{?prever:~%{prever}}"
    spec.update_tag("Version", "5.3.3")
    with spec.macro_definitions() as md:
        assert md.majorver.body == "1"
        assert md.minorver.body == "2"
        assert md.patchver.body == "3"
        assert md.prever.body == "alpha1"
        assert md.package_version.body == "4.0"
    assert spec.version == "5.3.3"
    spec.update_tag("Release", "2%{?dist}")
    assert spec.raw_release == "%{release}"
    with spec.macro_definitions() as md:
        assert md.release.body == "2%{?dist}"
    spec.update_tag(
        "Source0",
        "https://example.com/archived_releases/test/v6.0.0/test-v6.0.0.tar.xz",
    )
    assert spec.version == "6.0.0"
    with spec.sources() as sources:
        assert (
            sources[0].location
            == "https://example.com/archived_releases/%{name}/v%{version}/"
            "%{name}-v%{version}.tar.xz"
        )
    spec.update_tag(
        "Source0", "https://example.com/archived_releases/test-v7.2.1.tar.xz"
    )
    assert spec.version == "6.0.0"
    with spec.sources() as sources:
        assert (
            sources[0].location
            == "https://example.com/archived_releases/test-v7.2.1.tar.xz"
        )
    spec.update_tag("Source1", "tests-86.tar.xz")
    with spec.macro_definitions() as md:
        assert md.majorver.body == "1"
        assert md.minorver.body == "2"
    with spec.sources() as sources:
        assert sources[1].location == "tests-86.tar.xz"


def test_multiple_instances(spec_minimal, spec_autosetup):
    spec1 = Specfile(spec_minimal)
    spec2 = Specfile(spec_autosetup)
    spec1.version = "14.2"
    assert spec2.expanded_version == "0.1"
    with spec2.sources() as sources:
        assert sources[0].expanded_location == "test-0.1.tar.xz"
        sources.append("tests-%{version}.tar.xz")
        assert sources[1].expanded_location == "tests-0.1.tar.xz"


def test_includes(spec_includes):
    spec = Specfile(spec_includes)
    assert not spec.tainted
    with spec.patches() as patches:
        assert not patches
    assert spec.expand("%patches")
    with spec.tags() as tags:
        assert tags.provides.value.startswith("%(")
        for i in range(1, 4):
            assert f"test{i}-0.1" in tags.provides.expanded_value
    with spec.sections() as sections:
        assert sections.description[0] == "%include %{SOURCE3}"
        assert sections.description[1] == "%(cat %{S:4})"
    assert spec.parsed_sections.description[0] == "Test package"
    assert spec.parsed_sections.description[1] == "Additional description"
    for inc in ["patches.inc", "provides.inc", "description1.inc", "description2.inc"]:
        (spec.sourcedir / inc).unlink()
    with pytest.raises(RPMException):
        spec = Specfile(spec_includes)
    spec = Specfile(spec_includes, force_parse=True)
    assert spec.tainted
    with spec.patches() as patches:
        assert not patches
    assert not spec.expand("%patches")
    with spec.tags() as tags:
        assert tags.provides.value.startswith("%(")
    with spec.sections() as sections:
        assert sections.description[0] == "%include %{SOURCE3}"
        assert sections.description[1] == "%(cat %{S:4})"
    assert not "".join(spec.parsed_sections.description)
    for inc in ["macros1.inc", "macros2.inc"]:
        (spec.sourcedir / inc).unlink()
        with pytest.raises(RPMException):
            spec = Specfile(spec_includes, force_parse=True)
        assert not (spec.sourcedir / inc).is_file()


def test_shell_expansions(spec_shell_expansions):
    spec = Specfile(spec_shell_expansions)
    assert spec.expanded_version == "1035.4200"
    assert "C.UTF-8" in spec.expand("%numeric_locale")


def test_context_management(spec_autosetup, spec_traditional):
    spec = Specfile(spec_autosetup)
    with spec.tags() as tags:
        tags.license.value = "BSD"
        assert spec.license == "BSD"
        spec.license = "BSD-3-Clause"
        tags.patch0.value = "first_patch.patch"
        with spec.patches() as patches:
            assert patches[0].location == "first_patch.patch"
            patches[0].location = "patch_0.patch"
    assert spec.license == "BSD-3-Clause"
    with spec.patches() as patches:
        assert patches[0].location == "patch_0.patch"
    spec1 = Specfile(spec_autosetup)
    spec2 = Specfile(spec_traditional)
    with spec1.sections() as sections1, spec2.sections() as sections2:
        assert sections1 is not sections2
    with spec1.tags() as tags1, spec2.tags() as tags2:
        assert tags1 is not tags2
        assert tags1 == tags2


def test_copy(spec_autosetup):
    spec = Specfile(spec_autosetup)
    shallow_copy = copy.copy(spec)
    assert shallow_copy == spec
    assert shallow_copy is not spec
    assert shallow_copy._lines is spec._lines
    assert shallow_copy._parser is spec._parser
    deep_copy = copy.deepcopy(spec)
    assert deep_copy == spec
    assert deep_copy is not spec
    assert deep_copy._lines is not spec._lines
    assert deep_copy._parser is not spec._parser



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest

from specfile.utils import EVR, NEVR, NEVRA, get_filename_from_location


@pytest.mark.parametrize(
    "location, filename",
    [
        ("", ""),
        ("tarball-0.1.tar.gz", "tarball-0.1.tar.gz"),
        ("https://example.com", "example.com"),
        ("https://example.com#fragment", "example.com#fragment"),
        ("https://example.com/archive/tarball-0.1.tar.gz", "tarball-0.1.tar.gz"),
        (
            "https://example.com/archive/tarball-0.1.tar.gz#fragment",
            "tarball-0.1.tar.gz#fragment",
        ),
        (
            "https://example.com/download_tarball.cgi#/tarball-0.1.tar.gz",
            "tarball-0.1.tar.gz",
        ),
        (
            "https://example.com/tarball-latest.tar.gz#/file=tarball-0.1.tar.gz",
            "tarball-0.1.tar.gz",
        ),
    ],
)
def test_get_filename_from_location(location, filename):
    assert get_filename_from_location(location) == filename


@pytest.mark.parametrize(
    "evr, result",
    [
        ("0", EVR(version="0")),
        ("12.0-1", EVR(version="12.0", release="1")),
        ("2:56.8-5", EVR(epoch=2, version="56.8", release="5")),
        ("0.8.0-1.fc37", EVR(version="0.8.0", release="1.fc37")),
        ("0.5.0~rc2-1.el9", EVR(version="0.5.0~rc2", release="1.el9")),
        ("7.3-0.2.rc1.fc38", EVR(version="7.3", release="0.2.rc1.fc38")),
        (
            "7.3~rc1^20200701gdeadf00f-12.fc38",
            EVR(version="7.3~rc1^20200701gdeadf00f", release="12.fc38"),
        ),
    ],
)
def test_EVR_from_string(evr, result):
    assert EVR.from_string(evr) == result


@pytest.mark.parametrize(
    "nevr, result",
    [
        ("package-0", NEVR(name="package", version="0")),
        ("package-12.0-1", NEVR(name="package", version="12.0", release="1")),
        (
            "package-2:56.8-5",
            NEVR(name="package", epoch=2, version="56.8", release="5"),
        ),
        (
            "package-0.8.0-1.fc37",
            NEVR(name="package", version="0.8.0", release="1.fc37"),
        ),
        (
            "package-0.5.0~rc2-1.el9",
            NEVR(name="package", version="0.5.0~rc2", release="1.el9"),
        ),
        (
            "package-devel-7.3-0.2.rc1.fc38",
            NEVR(name="package-devel", version="7.3", release="0.2.rc1.fc38"),
        ),
        (
            "package-7.3~rc1^20200701gdeadf00f-12.fc38",
            NEVR(
                name="package", version="7.3~rc1^20200701gdeadf00f", release="12.fc38"
            ),
        ),
    ],
)
def test_NEVR_from_string(nevr, result):
    assert NEVR.from_string(nevr) == result


@pytest.mark.parametrize(
    "nevra, result",
    [
        (
            "package-12.0-1.x86_64",
            NEVRA(name="package", version="12.0", release="1", arch="x86_64"),
        ),
        (
            "package-2:56.8-5.aarch64",
            NEVRA(name="package", epoch=2, version="56.8", release="5", arch="aarch64"),
        ),
        (
            "package-0.8.0-1.fc37.armv6hl",
            NEVRA(name="package", version="0.8.0", release="1.fc37", arch="armv6hl"),
        ),
        (
            "package-0.5.0~rc2-1.el9.noarch",
            NEVRA(name="package", version="0.5.0~rc2", release="1.el9", arch="noarch"),
        ),
        (
            "package-devel-7.3-0.2.rc1.fc38.i686",
            NEVRA(
                name="package-devel", version="7.3", release="0.2.rc1.fc38", arch="i686"
            ),
        ),
        (
            "package-7.3~rc1^20200701gdeadf00f-12.fc38.riscv",
            NEVRA(
                name="package",
                version="7.3~rc1^20200701gdeadf00f",
                release="12.fc38",
                arch="riscv",
            ),
        ),
    ],
)
def test_NEVRA_from_string(nevra, result):
    assert NEVRA.from_string(nevra) == result



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import copy

import pytest
from flexmock import flexmock

from specfile.exceptions import SpecfileException
from specfile.sourcelist import Sourcelist, SourcelistEntry
from specfile.sources import ListSource, Patches, Sources, TagSource
from specfile.tags import Comments, Tag, Tags


@pytest.mark.parametrize(
    "tag_name, number",
    [
        ("Source", None),
        ("Source0", "0"),
        ("Patch0001", "0001"),
        ("Patch28", "28"),
        ("Patch99999", "99999"),
    ],
)
def test_tag_source_extract_number(tag_name, number):
    ts = TagSource(Tag(tag_name, "", "", Comments()))
    assert ts._extract_number() == number


@pytest.mark.parametrize(
    "tags, default, result",
    [
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
                ("Source2", "source2"),
            ],
            True,
            False,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source", "source0"),
                ("Source", "source1"),
                ("Source", "source2"),
            ],
            False,
            True,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source", "source0"),
            ],
            False,
            False,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source", "source0"),
            ],
            True,
            True,
        ),
    ],
)
def test_sources_detect_implicit_numbering(tags, default, result):
    sources = Sources(
        Tags([Tag(t, v, ": ", Comments()) for t, v in tags]),
        [],
        default_to_implicit_numbering=default,
    )
    assert sources._detect_implicit_numbering() == result


@pytest.mark.parametrize(
    "ref_name, ref_separator, number, name, separator",
    [
        ("Source", ": ", 28, "Source", ": "),
        ("Source0", ": ", 28, "Source28", ":"),
        ("Source0001", ":      ", 2, "Source0002", ":      "),
    ],
)
def test_sources_get_tag_format(ref_name, ref_separator, number, name, separator):
    sources = Sources(Tags(), [])
    reference = TagSource(Tag(ref_name, "", ref_separator, Comments()))
    assert sources._get_tag_format(reference, number) == (name, separator)


@pytest.mark.parametrize(
    "tags, number, index",
    [
        (["Name", "Version"], 0, 2),
        ([], 999, 0),
    ],
)
def test_sources_get_initial_tag_setup(tags, number, index):
    sources = Sources(Tags([Tag(t, "test", ": ", Comments()) for t in tags]), [])
    assert sources._get_initial_tag_setup(number) == (index, f"Source{number}", ": ")


@pytest.mark.parametrize(
    "tags, deduplicated_tags",
    [
        (["Source", "Source"], ["Source", "Source"]),
        (["Source0", "Source"], ["Source0", "Source"]),
        (
            ["Source0", "Source0", "Source", "Source"],
            ["Source0", "Source1", "Source", "Source"],
        ),
        (
            ["Source100", "Source100", "Source", "Source1028", "Source1011"],
            ["Source100", "Source101", "Source", "Source1028", "Source1011"],
        ),
        (
            ["Source2", "Source2", "Source4", "Source", "Source999"],
            ["Source2", "Source3", "Source4", "Source", "Source999"],
        ),
        (
            ["Source3", "Source4", "Source", "Source5"],
            ["Source3", "Source4", "Source", "Source6"],
        ),
        (
            [
                "Source",
                "Source",
                "Source100",
                "Source101",
                "Source101",
                "Source102",
                "Source",
            ],
            [
                "Source",
                "Source",
                "Source100",
                "Source101",
                "Source102",
                "Source103",
                "Source",
            ],
        ),
    ],
)
def test_sources_deduplicate_tag_names(tags, deduplicated_tags):
    sources = Sources(Tags([Tag(t, "test", ": ", Comments()) for t in tags]), [])
    sources._deduplicate_tag_names()
    assert [t.name for t in sources._tags] == deduplicated_tags


@pytest.mark.parametrize(
    "tags, sourcelists, index, location, number, cls",
    [
        ([("Name", "test"), ("Version", "0.1")], [], 0, "test", 0, TagSource),
        ([("Name", "test"), ("Version", "0.1")], [[]], 0, "test", 0, ListSource),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
            ],
            [],
            10,
            "test",
            2,
            TagSource,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
            ],
            [["source2", "source3"]],
            10,
            "test",
            4,
            ListSource,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
            ],
            [["source2", "source3"]],
            1,
            "test",
            1,
            TagSource,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source10", "source10"),
            ],
            [["source11", "source12"]],
            1,
            "test",
            10,
            TagSource,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
            ],
            [["source2", "source3"]],
            2,
            "test",
            2,
            ListSource,
        ),
        (
            [("Name", "test"), ("Version", "0.1"), ("Source0", "source0")],
            [],
            10,
            "source0",
            None,
            None,
        ),
    ],
)
def test_sources_insert(tags, sourcelists, index, location, number, cls):
    sources = Sources(
        Tags([Tag(t, v, ": ", Comments()) for t, v in tags]),
        [
            Sourcelist([SourcelistEntry(s, Comments()) for s in sl])
            for sl in sourcelists
        ],
    )
    if location in [v for t, v in tags if t.startswith(Sources.prefix)] + [
        s for sl in sourcelists for s in sl
    ]:
        with pytest.raises(SpecfileException):
            sources.insert(index, location)
    else:
        sources.insert(index, location)
        if index >= len(sources):
            index = len(sources) - 1
        assert isinstance(sources[index], cls)
        assert sources[index].number == number
        assert sources[index].location == location


@pytest.mark.parametrize(
    "tags, number, location, index",
    [
        ([("Name", "test"), ("Version", "0.1")], 28, "test", 0),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
                ("Source28", "source28"),
                ("Source999", "source999"),
            ],
            0,
            "test",
            0,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
                ("Source28", "source28"),
                ("Source999", "source999"),
            ],
            2,
            "test",
            2,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
                ("Source28", "source28"),
                ("Source999", "source999"),
            ],
            42,
            "test",
            3,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
                ("Source28", "source28"),
                ("Source999", "source999"),
            ],
            1000,
            "test",
            4,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source28", "source28"),
                ("Source42", "source42"),
                ("Source5", "source5"),
                ("Source", "source"),
            ],
            37,
            "test",
            1,
        ),
    ],
)
def test_sources_insert_numbered(tags, number, location, index):
    sources = Sources(Tags([Tag(t, v, ": ", Comments()) for t, v in tags]), [])
    if location in [v for t, v in tags if t.startswith(Sources.prefix)]:
        with pytest.raises(SpecfileException):
            sources.insert_numbered(number, location)
    else:
        assert sources.insert_numbered(number, location) == index
        assert isinstance(sources[index], TagSource)
        assert sources[index].number == number
        assert sources[index].location == location


@pytest.mark.parametrize(
    "tags, sourcelists, number, new_tags, new_sourcelists",
    [
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
            ],
            [],
            1,
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
            ],
            [],
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
            ],
            [],
            2,
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
            ],
            [],
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
            ],
            [["source2", "source3"]],
            2,
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
            ],
            [["source3"]],
        ),
    ],
)
def test_sources_remove_numbered(tags, sourcelists, number, new_tags, new_sourcelists):
    tags = Tags([Tag(t, v, ": ", Comments()) for t, v in tags])
    sourcelists = [
        Sourcelist([SourcelistEntry(s, Comments()) for s in sl]) for sl in sourcelists
    ]
    sources = Sources(tags, sourcelists)
    sources.remove_numbered(number)
    assert tags == Tags([Tag(t, v, ": ", Comments()) for t, v in new_tags])
    assert sourcelists == [
        Sourcelist([SourcelistEntry(s, Comments()) for s in sl])
        for sl in new_sourcelists
    ]


@pytest.mark.parametrize(
    "ref_name, ref_separator, number, name, separator",
    [
        ("Patch99", ":      ", 100, "Patch100", ":     "),
        ("Patch9999", ":  ", 28, "Patch28", ":    "),
        ("Patch0999", ":  ", 28, "Patch0028", ":  "),
        ("Source2", ":     ", 0, "Patch0", ":      "),
    ],
)
def test_patches_get_tag_format(ref_name, ref_separator, number, name, separator):
    patches = Patches(Tags(), [])
    reference = TagSource(Tag(ref_name, "", ref_separator, Comments()))
    assert patches._get_tag_format(reference, number) == (name, separator)


@pytest.mark.parametrize(
    "tags, number, index",
    [
        (["Name", "Version", "Source0"], 0, 3),
        (["Name", "Version", "Source0", "BuildRequires"], 1, 3),
        (["Name", "Version"], 2, 2),
        ([], 999, 0),
    ],
)
def test_patches_get_initial_tag_setup(tags, number, index):
    patches = Patches(Tags([Tag(t, "test", ": ", Comments()) for t in tags]), [])
    flexmock(patches).should_receive("_get_tag_format").and_return(
        f"Patch{number}", ": "
    )
    assert patches._get_initial_tag_setup(number) == (index, f"Patch{number}", ": ")


def test_copy_sources():
    sources = Sources(
        Tags([Tag("Name", "test", ": ", Comments())]),
        [
            Sourcelist([SourcelistEntry("%{name}-%{version}.tar.gz", Comments())]),
        ],
    )
    shallow_copy = copy.copy(sources)
    assert shallow_copy == sources
    assert shallow_copy is not sources
    assert shallow_copy._tags is sources._tags
    assert shallow_copy._sourcelists is sources._sourcelists
    deep_copy = copy.deepcopy(sources)
    assert deep_copy == sources
    assert deep_copy is not sources
    assert deep_copy._tags is not sources._tags
    assert deep_copy._sourcelists is not sources._sourcelists



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
import rpm
from flexmock import flexmock

from specfile.exceptions import MacroRemovalException
from specfile.macros import Macro, MacroLevel, Macros


def test_macros_parse():
    assert Macros._parse(
        [
            "-20= dump\t<builtin>\n",
            "-20: echo\t<builtin>\n",
            "-13: fontrpmname\t%{lua:\n",
            'local fonts = require "fedora.srpm.fonts"\n',
            'print(fonts.rpmname(rpm.expand("%1")))\n',
            "}\n",
            "-13: __scm_apply_git(qp:m:)\t\n",
            "%{__git} apply --index --reject %{-p:-p%{-p*}} -\n",
            '%{__git} commit %{-q} -m %{-m*} --author "%{__scm_author}"\n',
            "-13: py3_build_wheel\t%{expand:\\\n",
            '  CFLAGS="${CFLAGS:-${RPM_OPT_FLAGS}}" LDFLAGS="${LDFLAGS:-${RPM_LD_FLAGS}}"\\\n',
            "  %{__python3} %{py_setup} %{?py_setup_args} bdist_wheel %{?*}\n",
            "}\n",
            " -1: _sourcedir\t.\n",
            "========================\n",
        ]
    ) == [
        Macro("dump", None, "<builtin>", -20, True),
        Macro("echo", None, "<builtin>", -20, False),
        Macro(
            "fontrpmname",
            None,
            (
                "%{lua:\n"
                'local fonts = require "fedora.srpm.fonts"\n'
                'print(fonts.rpmname(rpm.expand("%1")))\n'
                "}"
            ),
            -13,
            False,
        ),
        Macro(
            "__scm_apply_git",
            "(qp:m:)",
            (
                "%{__git} apply --index --reject %{-p:-p%{-p*}} -\n"
                '%{__git} commit %{-q} -m %{-m*} --author "%{__scm_author}"'
            ),
            -13,
            False,
        ),
        Macro(
            "py3_build_wheel",
            None,
            (
                "%{expand:"
                '  CFLAGS="${CFLAGS:-${RPM_OPT_FLAGS}}" LDFLAGS="${LDFLAGS:-${RPM_LD_FLAGS}}"'
                "  %{__python3} %{py_setup} %{?py_setup_args} bdist_wheel %{?*}\n"
                "}"
            ),
            -13,
            False,
        ),
        Macro("_sourcedir", None, ".", -1, False),
    ]


def test_macros_remove():
    rpm.reloadConfig()
    macros = Macros.dump()
    rpm.addMacro("test", "1")
    rpm.addMacro("test", "2")
    rpm.addMacro("test", "3")
    assert set(Macros.dump()).difference(macros) == {
        Macro("test", None, "3", -1, False)
    }
    Macros.remove("test")
    Macros.remove("non_existent_macro")
    assert Macros.dump() == macros


def test_macros_remove_failure():
    # Expansion and removal of built-in macros is broken before rpm 4.17,
    # ensure that we are not stuck in an infinite loop
    rpm.reloadConfig()
    rpm.addMacro("foo", "bar")
    flexmock(rpm).should_receive("expandMacro").with_args("%foo").and_raise(rpm.error)
    with pytest.raises(MacroRemovalException):
        Macros.remove("foo")


def test_macros_define():
    rpm.reloadConfig()
    macros = Macros.dump()
    Macros.define("test", "1")
    assert set(Macros.dump()).difference(macros) == {
        Macro("test", None, "1", -1, False),
    }


def test_macros_reinit():
    Macros.reinit(MacroLevel.BUILTIN)
    assert all(m.level == MacroLevel.BUILTIN for m in Macros.dump())
